import numpy as np
from scipy.special import jv, jn_zeros, jnp_zeros

from .mode_basis import ModeBasis
from ..aperture import circular_aperture
from ..field import Field

def disk_harmonic(n, m, D=1, bc='dirichlet', grid=None):
	'''Create a disk harmonic.

	Parameters
	----------
	n : int
		Radial order
	m : int
		Azimuthal order
	D : scalar
		The diameter of the pupil.
	bc : string
		The boundary conditions to use. This can be either 'dirichlet', or
		'neumann' for a Dirichlet or Neumann boundary condition respectively.
	grid : Grid
		The grid on which to evaluate the function.

	Returns
	-------
	Field
		The disk harmonic function evaluated on `grid`.

	Raises
	------
	ValueError
		If the boundary condition is not recognized.
	'''
	polar_grid = grid.as_('polar')
	r = 2 * polar_grid.r / D
	theta = polar_grid.theta

	m_negative = m < 0
	m = abs(m)

	if bc == 'dirichlet':
		lambda_mn = jn_zeros(m, n)[-1]
		norm = 1
	elif bc == 'neumann':
		lambda_mn = jnp_zeros(m, n)[-1]
		norm = 1
	else:
		raise ValueError('Boundary condition not recognized.')

	if m_negative:
		z = norm * jv(m, lambda_mn * r) * np.sin(m * theta)
	else:
		z = norm * jv(m, lambda_mn * r) * np.cos(m * theta)

	# Do manual normalization for now...
	mask = circular_aperture(D)(grid) > 0.5
	norm = np.sqrt(np.sum(z[mask]**2))

	return Field(z * mask / norm, grid)

def disk_harmonic_energy(n, m, bc='dirichlet'):
	'''Get the energy of a disk harmonic function.

	This allows for functions to sort a disk harmonic mode basis on energy.

	Parameters
	----------
	n : int
		Radial order
	m : int
		Azimuthal order
	bc : string
		The boundary conditions to use. This can be either 'dirichlet', or
		'neumann' for a Dirichlet or Neumann boundary condition respectively.

	Returns
	-------
	scalar
		The energy corresponding to the mode.

	Raises
	------
	ValueError
		If the boundary condition is not recognized.
	'''
	m = abs(m)

	if bc == 'dirichlet':
		lambda_mn = jn_zeros(m, n)[-1]
	elif bc == 'neumann':
		lambda_mn = jnp_zeros(m, n)[-1]
	else:
		raise ValueError('Boundary condition not recognized.')

	return lambda_mn**2

def get_disk_harmonic_orders_sorted(num_modes, bc='dirichlet'):
	'''Return the orders for the first `num_modes` modes.

	Parameters
	----------
	num_modes : int
		The number of modes to return.
	bc : string
		The boundary conditions to use. This can be either 'dirichlet', or
		'neumann' for a Dirichlet or Neumann boundary condition respectively.

	Returns
	-------
	list of tuples
		A list of orders (n, m).
	'''
	orders = [(1,0)]
	energies = [disk_harmonic_energy(1, 0, bc)]

	results = []
	while len(results) < num_modes:
		k = np.argmin(energies)
		order = orders[k]

		if order[1] != 0:
			results.append((order[0], -order[1]))
		results.append(order)

		del orders[k]
		del energies[k]

		new_order = (order[0], order[1] + 1)
		if new_order not in results and new_order not in orders:
			orders.append(new_order)
			energies.append(disk_harmonic_energy(new_order[0], new_order[1], bc))
		new_order = (order[0] + 1, order[1])
		if new_order not in results and new_order not in orders:
			orders.append(new_order)
			energies.append(disk_harmonic_energy(new_order[0], new_order[1], bc))

	return results[:num_modes]

def make_disk_harmonic_basis(grid, num_modes, D=1, bc='dirichlet'):
	'''Create a disk harmonic mode basis.

	Parameters
	----------
	grid : Grid
		The grid on which to evaluate the disk harmonic modes.
	num_modes : int
		The number of modes to create.
	D : scalar
		The diameter of the disk.
	bc : string
		The boundary conditions to use. This can be either 'dirichlet', or
		'neumann' for a Dirichlet or Neumann boundary condition respectively.

	Returns
	-------
	ModeBasis
		The evaluated disk harmonic modes.
	'''
	orders = get_disk_harmonic_orders_sorted(num_modes, bc)

	modes = [disk_harmonic(order[0], order[1], D, bc, grid) for order in orders]
	return ModeBasis(modes, grid)



from hcipy import *
import numpy as np
import copy

def test_field_dot():
	grid = make_pupil_grid(2)

	a = np.random.randn(3, grid.size)
	A = np.random.randn(3, 3, grid.size)

	a = Field(a, grid)
	A = Field(A, grid)

	b = field_dot(A, a)
	bb = np.array([A[...,i].dot(a[...,i]) for i in range(grid.size)]).T

	assert np.allclose(b, bb)

	b = field_dot(a, a)
	bb = np.array([a[...,i].dot(a[...,i]) for i in range(grid.size)]).T

	assert np.allclose(b, bb)

	B = field_dot(A, A)
	BB = np.empty_like(B)
	for i in range(grid.size):
		BB[...,i] = A[...,i].dot(A[...,i])

	assert np.allclose(B, BB)

	b = field_dot(a, a)
	bb = np.array([a[...,i].dot(a[...,i]) for i in range(grid.size)])

	assert np.allclose(b, bb)

	n = np.random.randn(3)

	b = field_dot(A, n)
	bb = np.array([A[...,i].dot(n) for i in range(grid.size)]).T

	assert np.allclose(b, bb)

	b = field_dot(n, A)
	bb = np.array([n.dot(A[...,i]) for i in range(grid.size)]).T

	assert np.allclose(b, bb)

	N = np.random.randn(3,3)

	B = field_dot(A, N)
	BB = np.empty_like(B)
	for i in range(grid.size):
		BB[...,i] = A[...,i].dot(N)

	assert np.allclose(B, BB)

def test_field_trace():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3,3,grid.size), grid)

	B = field_trace(A)
	BB = np.array([np.trace(A[...,i]) for i in range(grid.size)])

	assert np.allclose(B, BB)

def test_field_inv():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3,3,grid.size), grid)

	B = field_inv(A)
	BB = np.empty_like(B)
	for i in range(grid.size):
		BB[...,i] = np.linalg.inv(A[...,i])

	assert np.allclose(B, BB)

def test_field_transpose():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3, 3, grid.size), grid)

	B = field_transpose(A)
	BB = np.empty_like(B)
	for i in range(grid.size):
		BB[..., i] = A[..., i].T

	assert np.allclose(B, BB)

def test_field_conjugate_transpose():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3, 3, grid.size), grid)

	B = field_conjugate_transpose(A)
	BB = np.empty_like(B)
	for i in range(grid.size):
		BB[..., i] = A[..., i].T.conj()

	assert np.allclose(B, BB)

def test_field_adjoint():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3, 3, grid.size), grid)

	B = field_adjoint(A)
	BB = np.empty_like(B)
	for i in range(grid.size):
		BB[..., i] = np.linalg.inv(A[..., i]) * np.linalg.det(A[..., i])

	assert np.allclose(B, BB)

def test_field_inverse_tikhonov():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3,3,grid.size), grid)

	for reg in [1e-1, 1e-3, 1e-6]:
		B = field_inverse_tikhonov(A, reg)
		BB = np.empty_like(B)

		for i in range(grid.size):
			BB[...,i] = inverse_tikhonov(A[...,i], reg)

		assert np.allclose(B, BB)

def test_field_inverse_truncated():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3,3,grid.size), grid)

	for reg in [1e-1, 1e-3, 1e-6]:
		B = field_inverse_truncated(A, reg)
		BB = np.empty_like(B)

		for i in range(grid.size):
			BB[...,i] = inverse_truncated(A[...,i], reg)

		assert np.allclose(B, BB)

def test_field_inverse_truncated_modal():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3,3,grid.size), grid)

	for num_modes in [1, 2]:
		B = field_inverse_truncated_modal(A, num_modes)
		BB = np.empty_like(B)

		for i in range(grid.size):
			BB[...,i] = inverse_truncated_modal(A[...,i], num_modes)

		assert np.allclose(B, BB)

def test_field_cross():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(3, grid.size), grid)
	B = Field(np.random.randn(3, grid.size), grid)

	C = field_cross(A, B)
	CC = np.empty_like(C)
	for i in range(grid.size):
		CC[..., i] = np.cross(A[:, i], B[:, i])

	assert np.allclose(C, CC)

def test_field_svd():
	grid = make_pupil_grid(2)

	A = Field(np.random.randn(5,10,grid.size), grid)

	U, S, Vh = field_svd(A)
	u, s, vh = field_svd(A, False)

	for i in range(grid.size):
		svd = np.linalg.svd(A[...,i])

		assert np.allclose(U[...,i], svd[0])
		assert np.allclose(S[...,i], svd[1])
		assert np.allclose(Vh[...,i], svd[2])

		svd2 = np.linalg.svd(A[...,i], full_matrices=False)

		assert np.allclose(u[...,i], svd2[0])
		assert np.allclose(s[...,i], svd2[1])
		assert np.allclose(vh[...,i], svd2[2])

def test_grid_hashing():
	grid1 = make_pupil_grid(128)

	grid2 = CartesianGrid(SeparatedCoords(copy.deepcopy(grid1.separated_coords)))
	assert hash(grid1) != hash(grid2)

	grid3 = CartesianGrid(UnstructuredCoords(copy.deepcopy(grid1.coords)))
	assert hash(grid1) != hash(grid3)

	grid4 = make_pupil_grid(128)
	assert hash(grid1) == hash(grid4)

	grid5 = PolarGrid(grid1.coords)
	assert hash(grid1) != hash(grid5)

	grid6 = CartesianGrid(copy.deepcopy(grid1.coords))
	assert hash(grid1) == hash(grid6)

	grid7 = grid1.scaled(2)
	assert hash(grid1) != hash(grid7)

	grid8 = grid1.scaled(2)
	assert hash(grid1) != hash(grid8)
	assert hash(grid7) == hash(grid8)

	grid9 = make_pupil_grid(256)
	assert hash(grid1) != hash(grid9)

def test_grid_supersampled():
	g = make_uniform_grid(128, [1, 1])
	g2 = make_supersampled_grid(g, 4)
	g3 = make_subsampled_grid(g2, 4)

	assert np.allclose(g.x, g3.x)
	assert np.allclose(g.y, g3.y)

	g4 = make_supersampled_grid(make_supersampled_grid(g, 2), 2)

	assert np.allclose(g2.x, g4.x)
	assert np.allclose(g2.y, g4.y)



import numpy as np
import scipy.sparse

from ..mode_basis import ModeBasis
from .deformable_mirror import DeformableMirror

class SegmentedDeformableMirror(DeformableMirror):
	'''A segmented deformable mirror.

	This deformable mirror class can simulate devices such as those
	made by IrisAO and BMC. All segments are controlled in piston,
	tip and tilt.

	Parameters
	----------
	segments : ModeBasis
		A mode basis with all segments.
	'''
	def __init__(self, segments):
		self.segments = segments
		self.actuators = np.zeros(len(segments) * 3)
		self.input_grid = segments.grid

	@property
	def segments(self):
		'''The segments of this deformable mirror in a ModeBasis.
		'''
		return self._segments

	@segments.setter
	def segments(self, segments):
		self._segments = segments

		tip = []
		tilt = []

		for segment in segments:
			segment_mean = segment.mean()
			norm = np.mean(segment**2) - segment_mean**2

			tip_mode = segment * segment.grid.x
			if norm != 0:
				beta = ((tip_mode * segment).mean() - segment_mean * tip_mode.mean()) / norm
				tip_mode -= beta * segment

			tip_mode = scipy.sparse.csr_matrix(tip_mode)
			tip_mode.eliminate_zeros()
			tip.append(tip_mode)

			tilt_mode = segment * segment.grid.y
			if norm != 0:
				beta = ((tilt_mode * segment).mean() - segment_mean * tilt_mode.mean()) / norm
				tilt_mode -= beta * segment

			tilt_mode = scipy.sparse.csr_matrix(tilt_mode)
			tilt_mode.eliminate_zeros()
			tilt.append(tilt_mode)

		tip = ModeBasis(tip)
		tilt = ModeBasis(tilt)

		self.influence_functions = segments + tip + tilt

	def get_segment_actuators(self, segment_id):
		'''Get the actuators for an individual segment of the DM.

		Parameters
		----------
		segment_id : int
			The index of the segment for which to get the actuators.

		Returns
		-------
		piston : scalar
			The piston of the segment in meters.
		tip : scalar
			The tip of the segment in radians.
		tilt : scalar
			The tilt of the segment in radians.
		'''
		piston = self.actuators[segment_id]
		tip = self.actuators[segment_id + len(self._segments)]
		tilt = self.actuators[segment_id + 2 * len(self._segments)]

		return (piston, tip, tilt)

	def set_segment_actuators(self, segment_id, piston, tip, tilt):
		'''Set the actuators for an individual segment of the DM.

		Parameters
		----------
		segment_id : int
			The index of the segment for which to get the actuators.
		piston : scalar
			The piston of the segment in meters.
		tip : scalar
			The tip of the segment in radians.
		tilt : scalar
			The tilt of the segment in radians.
		'''
		self.actuators[segment_id] = piston
		self.actuators[segment_id + len(self._segments)] = tip
		self.actuators[segment_id + 2 * len(self._segments)] = tilt



def get_version():
	if get_version._version is None:
		from pkg_resources import get_distribution, DistributionNotFound

		try:
			get_version._version = get_distribution('hcipy').version
		except DistributionNotFound:
			# package is not installed
			pass

	return get_version._version

get_version._version = None



import numpy as np
import os
from hcipy import *

def test_grid_io():
	grid1 = make_pupil_grid(128)
	grid2 = CartesianGrid(SeparatedCoords(grid1.separated_coords))
	grid3 = grid1.as_('polar')

	grids = [grid1, grid2, grid3]
	formats = ['asdf', 'fits', 'fits.gz', 'pkl', 'pickle']
	filenames = ['grid_test.' + fmt for fmt in formats]

	for g in grids:
		for fname in filenames:
			write_grid(g, fname)
			new_grid = read_grid(fname)

			assert hash(g) == hash(new_grid)

			os.remove(fname)

def test_field_io():
	grid = make_pupil_grid(128)
	field = circular_aperture(1)(grid)

	formats = ['asdf', 'fits', 'fits.gz', 'pkl', 'pickle']
	filenames = ['field_test.' + fmt for fmt in formats]

	for fname in filenames:
		write_field(field, fname)
		new_field = read_field(fname)

		assert np.allclose(field, new_field)
		assert hash(field.grid) == hash(new_field.grid)

		os.remove(fname)

def test_mode_basis_io():
	grid = make_pupil_grid(128)
	mode_bases = [
		make_zernike_basis(20, 1, grid, 1),
		make_xinetics_influence_functions(grid, 8, 1 / 8)
	]

	formats = ['asdf', 'fits', 'fits.gz', 'pkl', 'pickle']
	filenames = ['mode_basis_test.' + fmt for fmt in formats]

	for mode_basis in mode_bases:
		for fname in filenames:
			write_mode_basis(mode_basis, fname)
			new_mode_basis = read_mode_basis(fname)

			assert hash(new_mode_basis.grid) == hash(mode_basis.grid)
			assert new_mode_basis.is_sparse == mode_basis.is_sparse

			for i in range(mode_basis.num_modes):
				assert np.allclose(mode_basis[i], new_mode_basis[i])

			os.remove(fname)



import numpy as np
import matplotlib.pyplot as plt
from hcipy import *
import pytest
import os

def test_agnostic_apodizer():
	wavelength = 0.2
	aperture_achromatic = circular_aperture(1)
	aperture_chromatic = lambda input_grid, wavelength: circular_aperture(wavelength)(input_grid)

	apod_achromatic = Apodizer(aperture_achromatic)
	apod_chromatic = Apodizer(aperture_chromatic)

	phase_chromatic = lambda input_grid, wavelength: zernike(2, 0)(input_grid) * wavelength
	apod_phase_chromatic = PhaseApodizer(phase_chromatic)

	phase_achromatic = zernike(2, 0)
	apod_phase_achromatic = PhaseApodizer(phase_achromatic)

	filter_curve = lambda wavelength: np.sqrt(wavelength)
	apod_filter = Apodizer(filter_curve)

	for wl in [0.2, 0.4, 0.7]:
		for num_pix in [32, 64, 128]:
			for diameter in [0.5, 1, 2]:
				pupil_grid = make_pupil_grid(num_pix, diameter)
				wf = Wavefront(pupil_grid.ones(), wl)

				pup = apod_achromatic(wf).electric_field
				assert np.allclose(pup, aperture_achromatic(pupil_grid))

				pup = apod_chromatic(wf).electric_field
				assert np.allclose(pup, aperture_chromatic(pupil_grid, wl))

				pup = apod_phase_achromatic(wf).electric_field
				assert np.allclose(pup, np.exp(1j * phase_achromatic(pupil_grid)))

				pup = apod_phase_chromatic(wf).electric_field
				assert np.allclose(pup, np.exp(1j * phase_chromatic(pupil_grid, wl)))

				pup = apod_filter(wf).electric_field
				assert np.allclose(pup, filter_curve(wl))

def test_statistics_noisy_detector():
	N = 256
	grid = make_pupil_grid(N)

	field = Field(np.ones(N**2), grid)

	#First we test photon noise, dark current noise and read noise.
	flat_field = 0
	dark_currents = np.logspace(1, 6, 6)
	read_noises = np.logspace(1, 6, 6)
	photon_noise = True

	for dc in dark_currents:
		for rn in read_noises:
			#The test detector.
			detector = NoisyDetector(detector_grid=grid, include_photon_noise=photon_noise, flat_field=flat_field, dark_current_rate=dc, read_noise=rn)

			#The integration times we will test.
			integration_time = np.logspace(1,6,6)

			for t in integration_time:
				# integration
				detector.integrate(field, t)

				# read out
				measurement = detector.read_out()

				# The std of the data by the detector.
				std_measurement = np.std(measurement)

				# The std that we expect given the input.
				expected_std = np.sqrt(field[0] * t + rn**2 + dc * t)

				assert np.isclose(expected_std, std_measurement, rtol=2e-02, atol=1e-05)

	#Now we test the flat field separately.
	flat_fields = np.linspace(0, 1, 100)
	dark_current = 0
	read_noise = 0
	photon_noise = False

	for ff in flat_fields:
		detector = NoisyDetector(detector_grid=grid, include_photon_noise=photon_noise, flat_field=ff, dark_current_rate=dark_current, read_noise=read_noise)

		integration_times = np.logspace(1, 6, 6)

		for t in integration_times:
			# integration and read out
			detector.integrate(field, t)
			measurement = detector.read_out()

			# The std of the data by the detector.
			std_measurement = np.std(measurement)

			# The std that we expect given the input.
			expected_std = ff * field[0] * t

			assert np.isclose(expected_std, std_measurement, rtol=2e-02, atol=1e-05)

	aperture = circular_aperture(1)(grid)
	subsamplings = [1, 2, 4, 8]

	for subsampling in subsamplings:
		grid_subsampled = make_subsampled_grid(grid, subsampling)
		aperture_subsampled = subsample_field(aperture, subsampling, grid_subsampled, statistic='sum')

		# the detector with the new subsampling factor
		detector = NoisyDetector(detector_grid=grid_subsampled, include_photon_noise=False, subsampling=subsampling)

		# integrating and reading out
		detector.integrate(aperture, dt=1)
		detector_image = detector.read_out()

		# testing if the image from the detector matches the subsampled aperture
		assert np.allclose(aperture_subsampled, detector_image)

def test_glass_catalogue():
	bk7 = get_refractive_index('N-BK7')
	assert np.allclose(bk7(500e-9), 1.5214144761028994)

	with pytest.raises(ValueError) as exception_info:
		get_refractive_index('N-Bk7')
	assert 'Did you mean' in str(exception_info.value)

	with pytest.raises(ValueError) as exception_info:
		get_refractive_index('N-Ba7')
	assert 'Did you mean' not in str(exception_info.value)

def test_deformable_mirror():
	num_pix = 256
	x_tilt = np.radians(15)
	y_tilt = np.radians(10)
	z_tilt = np.radians(20)
	grid = make_pupil_grid(num_pix, 1.5)

	functions = [make_gaussian_influence_functions, make_xinetics_influence_functions]

	for num_actuators_across_pupil in [12, 16]:
		actuator_spacing = 1 / num_actuators_across_pupil
		num_actuators = num_actuators_across_pupil**2

		# Check if the actuator spacing of an unrotated DM is correct
		actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing)
		assert np.allclose(actuator_positions.delta, actuator_spacing)

		actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing, x_tilt=x_tilt, y_tilt=y_tilt, z_tilt=z_tilt)

		for func in functions:
			influence_functions = func(grid, num_actuators_across_pupil, actuator_spacing, x_tilt=x_tilt, y_tilt=y_tilt, z_tilt=z_tilt)

			deformable_mirror = DeformableMirror(influence_functions)

			# Check number of generated influence functions
			assert len(influence_functions) == num_actuators
			assert deformable_mirror.num_actuators == num_actuators

			# Check that influence functions are sparse
			assert influence_functions.is_sparse

			# Check if the centroids for each influence function are close to the predicted position.
			for act, p in zip(influence_functions, actuator_positions.points):
				x_pos = (act * grid.x).sum() / act.sum()
				y_pos = (act * grid.y).sum() / act.sum()

				assert np.allclose([x_pos, y_pos], p, atol=0.02)

			# Mirror should start out flattened
			assert np.std(deformable_mirror.surface) < 1e-12

			for i in np.random.randint(0, num_actuators, size=10):
				deformable_mirror.actuators[i] = np.random.randn(1)

				# Mirror should not be flat with poked actuator
				assert np.std(deformable_mirror.surface) > 1e-12

				deformable_mirror.actuators[i] = 0

			# Mirror should be flat again
			assert np.std(deformable_mirror.surface) < 1e-12

			deformable_mirror.random(0.001)

			# Mirror should have random actuator pokes
			assert np.std(deformable_mirror.surface) > 1e-12

			wf = Wavefront(grid.ones())
			wf_out = deformable_mirror.forward(wf)

			# Check that the correct phase is applied
			assert np.allclose(wf_out.phase, deformable_mirror.phase_for(1))

			wf_in = deformable_mirror.backward(wf_out)

			# Check that the deformable mirror is phase only
			assert np.allclose(wf_in.electric_field, wf.electric_field)

			# Check OPD
			assert np.allclose(deformable_mirror.opd, 2 * deformable_mirror.surface)

			deformable_mirror.flatten()

			# Check that the deformable mirror is flat again
			assert np.std(deformable_mirror.surface) < 1e-12

def test_segmented_deformable_mirror():
	num_pix = 256
	grid = make_pupil_grid(num_pix)

	for num_rings in [2, 4, 5]:
		num_segments_expected = 3 * (num_rings + 1) * num_rings + 1

		segment_positions = make_hexagonal_grid(0.5 / num_rings * np.sqrt(3) / 2, num_rings)
		aperture, segments = make_segmented_aperture(hexagonal_aperture(0.5 / num_rings - 0.003, np.pi / 2), segment_positions, return_segments=True)

		aperture = evaluate_supersampled(aperture, grid, 2)
		segments = evaluate_supersampled(segments, grid, 2)

		# Check number of generated segments
		assert len(segments) == num_segments_expected

		segmented_mirror = SegmentedDeformableMirror(segments)

		# Mirror should start out flattened.
		assert np.std(segmented_mirror.surface) < 1e-12

		for i in np.random.randint(0, num_segments_expected, size=10):
			piston = np.random.randn(1)

			while np.abs(piston) < 1e-5:
				piston = np.random.randn(1)

			segmented_mirror.set_segment_actuators(i, piston, 0, 0)

			# Mirror should have the correct piston
			assert np.abs(segmented_mirror.get_segment_actuators(i)[0] - piston) / piston < 1e-5
			assert np.abs(np.ptp(segmented_mirror.surface) - piston) / piston < 1e-5

			tip, tilt = np.random.randn(2)

			while np.abs(tip) < 1e-5:
				tip = np.random.randn(1)

			while np.abs(tilt) < 1e-5:
				tilt = np.random.randn(1)

			segmented_mirror.set_segment_actuators(i, 0, tip, tilt)

			# Mirror should be distorted with tip and tilt on a segment
			assert np.std(segmented_mirror.surface > 1e-12)

			# Return segment to zero position
			segmented_mirror.set_segment_actuators(i, 0, 0, 0)

		# Run labeling of segments
		imshow_field(aperture)
		label_actuator_centroid_positions(segments)
		plt.clf()

def test_wavefront_stokes():
	N = 4
	grid = make_pupil_grid(N)

	stokes = np.random.uniform(-1,1,4)
	stokes[0] = np.abs(stokes[0])+np.linalg.norm(stokes[1:])

	amplitude = np.random.uniform(0, 1, (2,2,N**2))
	phase = np.random.uniform(0, 2 * np.pi, (2,2,N**2))

	jones_field = Field(amplitude * np.exp(1j * phase), grid)

	determinants = field_determinant(jones_field)
	jones_field[:, :, determinants > 1] *= np.random.uniform(0, 1, np.sum(determinants > 1)) / determinants[determinants > 1]

	jones_element = JonesMatrixOpticalElement(jones_field)

	mueller_field = jones_to_mueller(jones_field)

	field = Field(np.ones(N**2), grid)
	stokes_wavefront = Wavefront(field, input_stokes_vector=stokes)

	jones_element_forward = jones_element.forward(stokes_wavefront)
	mueller_forward = field_dot(mueller_field, stokes)

	assert np.allclose(jones_element_forward.I, mueller_forward[0])
	assert np.allclose(jones_element_forward.Q, mueller_forward[1])
	assert np.allclose(jones_element_forward.U, mueller_forward[2])
	assert np.allclose(jones_element_forward.V, mueller_forward[3])

def test_degree_and_angle_of_polarization():
	grid = make_pupil_grid(16)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, 0])
	assert np.allclose(wf.degree_of_polarization, 0)
	assert np.allclose(wf.degree_of_linear_polarization, 0)
	assert np.allclose(wf.degree_of_circular_polarization, 0)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 1, 0, 0])
	assert np.allclose(wf.degree_of_polarization, 1)
	assert np.allclose(wf.degree_of_linear_polarization, 1)
	assert np.allclose(wf.angle_of_linear_polarization, 0)
	assert np.allclose(wf.degree_of_circular_polarization, 0)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0.5, 0, 0])
	assert np.allclose(wf.degree_of_polarization, 0.5)
	assert np.allclose(wf.degree_of_linear_polarization, 0.5)
	assert np.allclose(wf.angle_of_linear_polarization, 0)
	assert np.allclose(wf.degree_of_circular_polarization, 0)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, -1, 0, 0])
	assert np.allclose(wf.degree_of_polarization, 1)
	assert np.allclose(wf.degree_of_linear_polarization, 1)
	assert np.allclose(wf.angle_of_linear_polarization, np.pi / 2)
	assert np.allclose(wf.degree_of_circular_polarization, 0)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 1, 0])
	assert np.allclose(wf.degree_of_polarization, 1)
	assert np.allclose(wf.degree_of_linear_polarization, 1)
	assert np.allclose(wf.angle_of_linear_polarization, np.pi / 4)
	assert np.allclose(wf.degree_of_circular_polarization, 0)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, 1])
	assert np.allclose(wf.degree_of_polarization, 1)
	assert np.allclose(wf.degree_of_linear_polarization, 0)
	assert np.allclose(wf.degree_of_circular_polarization, 1)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, 0.5])
	assert np.allclose(wf.degree_of_polarization, 0.5)
	assert np.allclose(wf.degree_of_linear_polarization, 0)
	assert np.allclose(wf.degree_of_circular_polarization, 0.5)

	wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, -1])
	assert np.allclose(wf.degree_of_polarization, 1)
	assert np.allclose(wf.degree_of_linear_polarization, 0)
	assert np.allclose(wf.degree_of_circular_polarization, -1)

	wf = Wavefront(grid.ones(), input_stokes_vector=[2, np.sqrt(2), 0, np.sqrt(2)])
	assert np.allclose(wf.degree_of_polarization, 1)
	assert np.allclose(wf.degree_of_linear_polarization, 1 / np.sqrt(2))
	assert np.allclose(wf.degree_of_circular_polarization, 1 / np.sqrt(2))

def mueller_matrix_for_general_linear_retarder(theta, delta):
	'''Analytic expression Mueller matrix linear retarder.

	Parameters
	----------
	theta : scaler
		rotation angle optic in radians
	delta : scalar
		retardance optic in radians
	'''
	retarder = np.zeros((4, 4))

	retarder[0, 0] = 1

	retarder[1, 1] = np.cos(2 * theta)**2 + np.sin(2 * theta)**2 * np.cos(delta)
	retarder[1, 2] = np.cos(2 * theta) * np.sin(2*theta) * (1 - np.cos(delta))
	retarder[1, 3] = np.sin(2 * theta) * np.sin(delta)

	retarder[2, 1] = np.cos(2 * theta) * np.sin(2 * theta) * (1 - np.cos(delta))
	retarder[2, 2] = np.cos(2 * theta)**2 * np.cos(delta) + np.sin(2 * theta)**2
	retarder[2, 3] = -np.cos(2 * theta) * np.sin(delta)

	retarder[3, 1] = -np.sin(2 * theta) * np.sin(delta)
	retarder[3, 2] = np.cos(2 * theta) * np.sin(delta)
	retarder[3, 3] = np.cos(delta)

	return retarder

def mueller_matrix_for_general_linear_polarizer(theta):
	'''Analytic expression Mueller matrix linear polarizer.

	Parameters
	----------
	theta : scaler
		rotation angle optic in radians
	'''
	polarizer = np.zeros((4, 4))

	polarizer[0, 0] = 1
	polarizer[0, 1] = np.cos(2 * theta)
	polarizer[0, 2] = np.sin(2 * theta)

	polarizer[1, 0] = np.cos(2 * theta)
	polarizer[1, 1] = np.cos(2 * theta)**2
	polarizer[1, 2] = 0.5 * np.sin(4*theta)

	polarizer[2, 0] = np.sin(2 * theta)
	polarizer[2, 1] = 0.5 * np.sin(4 * theta)
	polarizer[2, 2] = np.sin(2 * theta)**2

	polarizer *= 0.5
	return polarizer

def test_polarization_elements():
	N = 1
	grid = make_pupil_grid(N)
	test_field = grid.ones()

	# Stokes vectors used for testing.
	stokes_vectors = [
		None,
		np.array([1, 0, 0, 0]), # unpolarized
		np.array([1, 1, 0, 0]), # +Q polarized
		np.array([1, -1, 0, 0]), # -Q polarized
		np.array([1, 0, 1, 0]), # +U polarized
		np.array([1, 0, -1, 0]), # -U polarized
		np.array([1, 0, 0, 1]), # +V polarized
		np.array([1, 0, 0, -1])] # -V polarized

	# Angles of the optics that will be tested.
	angles = [-45, -22.5, 0, 22.5, 45, 90] # degrees

	# Test QWPs, HWPs, polarizers and PBSs
	for stokes_vector in stokes_vectors:
		test_wf = Wavefront(test_field, input_stokes_vector=stokes_vector)

		if stokes_vector is None:
			# Set Stokes vector for futher calculation to unpolarized light.
			stokes_vector = np.array([1, 0, 0, 0])

		for angle in angles:
			# Create quarterwave plate.
			QWP_hcipy = QuarterWavePlate(np.radians(angle))

			# Test if Mueller matrix is the same as reference.
			QWP_ref = mueller_matrix_for_general_linear_retarder(np.radians(angle), -np.pi / 2)
			assert np.allclose(QWP_hcipy.mueller_matrix, QWP_ref)

			# Propagate wavefront through optical element.
			wf_forward_QWP = QWP_hcipy.forward(test_wf)

			# Test if result is the same as reference.
			reference_stokes_post_QWP = field_dot(QWP_ref, stokes_vector)
			assert np.allclose(wf_forward_QWP.stokes_vector[:, 0], reference_stokes_post_QWP)

			# Test backward propagation.
			wf_forward_backward_QWP = QWP_hcipy.backward(wf_forward_QWP)
			assert np.allclose(wf_forward_backward_QWP.stokes_vector[:, 0], stokes_vector)

			# Test power conservation.
			assert np.allclose(wf_forward_QWP.I, test_wf.I)

			# Create halfwave plate
			HWP_hcipy = HalfWavePlate(np.radians(angle))

			# Test if Mueller matrix is the same as reference.
			HWP_ref = mueller_matrix_for_general_linear_retarder(np.radians(angle), -np.pi)
			assert np.allclose(HWP_hcipy.mueller_matrix, HWP_ref)

			# Propagate wavefront through optical element.
			wf_forward_HWP = HWP_hcipy.forward(test_wf)

			# Test if result is the same as reference.
			reference_stokes_post_HWP = field_dot(HWP_ref, stokes_vector)
			assert np.allclose(wf_forward_HWP.stokes_vector[:, 0], reference_stokes_post_HWP)

			# Test backward propagation.
			wf_forward_backward_HWP = HWP_hcipy.backward(wf_forward_HWP)
			assert np.allclose(wf_forward_backward_HWP.stokes_vector[:,0], stokes_vector)

			# Test power conservation.
			assert np.allclose(wf_forward_HWP.I, test_wf.I)

			# Create polarizer.
			polarizer_hcipy = LinearPolarizer(np.radians(angle))

			# Test if Mueller matrix is the same as reference.
			polarizer_ref = mueller_matrix_for_general_linear_polarizer(np.radians(angle))
			assert np.allclose(polarizer_hcipy.mueller_matrix, polarizer_ref)

			# Propagate wavefront through optical element.
			wf_forward_polarizer = polarizer_hcipy.forward(test_wf)

			# Test if result is the same as reference.
			reference_stokes_post_polarizer = field_dot(polarizer_ref, stokes_vector)
			assert np.allclose(wf_forward_polarizer.stokes_vector[:, 0], reference_stokes_post_polarizer)

			# Test backward propagation.
			wf_backward_polarizer = polarizer_hcipy.backward(test_wf)
			assert np.allclose(wf_backward_polarizer.stokes_vector[:, 0], reference_stokes_post_polarizer)

			# Create polarizing beam splitter
			LPBS_hcipy = LinearPolarizingBeamSplitter(np.radians(angle))

			# Test if Mueller matrices are the same as reference.
			polarizer_1_ref = mueller_matrix_for_general_linear_polarizer(np.radians(angle))
			polarizer_2_ref = mueller_matrix_for_general_linear_polarizer(np.radians(angle + 90))
			assert np.allclose(LPBS_hcipy.mueller_matrices[0], polarizer_1_ref)
			assert np.allclose(LPBS_hcipy.mueller_matrices[1], polarizer_2_ref)

			# Propagate wavefront through optical element.
			wf_forward_polarizer_1, wf_forward_polarizer_2 = LPBS_hcipy.forward(test_wf)

			# Test if result is the same as reference.
			reference_stokes_post_polarizer_1 = field_dot(polarizer_1_ref, stokes_vector)
			reference_stokes_post_polarizer_2 = field_dot(polarizer_2_ref, stokes_vector)
			assert np.allclose(wf_forward_polarizer_1.stokes_vector[:, 0], reference_stokes_post_polarizer_1)
			assert np.allclose(wf_forward_polarizer_2.stokes_vector[:, 0], reference_stokes_post_polarizer_2)

			# Test power conservation.
			assert np.allclose(test_wf.I, wf_forward_polarizer_1.I + wf_forward_polarizer_2.I)

			# Test multiplication of polarization optics
			# 1) JonesMatrixOpticalElement * JonesMatrixOpticalElement
			multiplication_test_1 = polarizer_hcipy * QWP_hcipy
			# 2) JonesMatrixOpticalElement * numpy array
			multiplication_test_2 = polarizer_hcipy * QWP_hcipy.jones_matrix

			multiplication_test_ref = np.dot(polarizer_ref, QWP_ref)

			# testing if the Mueller matrices are the same
			assert np.allclose(multiplication_test_1.mueller_matrix, multiplication_test_ref)
			assert np.allclose(multiplication_test_2.mueller_matrix, multiplication_test_ref)

			# propagating the wavefront through the optics
			wf_forward_multiplication_1 = multiplication_test_1.forward(test_wf)
			wf_forward_multiplication_2 = multiplication_test_2.forward(test_wf)

			reference_stokes_post_multiplication = field_dot(multiplication_test_ref, stokes_vector)

			assert np.allclose(wf_forward_multiplication_1.stokes_vector[:, 0], reference_stokes_post_multiplication)
			assert np.allclose(wf_forward_multiplication_2.stokes_vector[:, 0], reference_stokes_post_multiplication)

		# Create polarizing beam splitter
		CPBS_hcipy = CircularPolarizingBeamSplitter()

		# Test if Mueller matrices are the same as reference.
		circ_polarizer_1_ref = mueller_matrix_for_general_linear_polarizer(0)
		circ_polarizer_2_ref = mueller_matrix_for_general_linear_polarizer(np.radians(90))
		QWP_1_ref = mueller_matrix_for_general_linear_retarder(np.pi/4, -np.pi / 2)
		CP_1_ref = np.dot(circ_polarizer_1_ref,QWP_1_ref)
		CP_2_ref = np.dot(circ_polarizer_2_ref,QWP_1_ref)

		assert np.allclose(CPBS_hcipy.mueller_matrices[0], CP_1_ref)
		assert np.allclose(CPBS_hcipy.mueller_matrices[1], CP_2_ref)

		# Propagate wavefront through optical element.
		wf_forward_circ_polarizer_1, wf_forward_circ_polarizer_2 = CPBS_hcipy.forward(test_wf)

		# Test if result is the same as reference.
		reference_stokes_post_circ_polarizer_1 = field_dot(CP_1_ref, stokes_vector)
		reference_stokes_post_circ_polarizer_2 = field_dot(CP_2_ref, stokes_vector)
		assert np.allclose(wf_forward_circ_polarizer_1.stokes_vector[:, 0], reference_stokes_post_circ_polarizer_1)
		assert np.allclose(wf_forward_circ_polarizer_2.stokes_vector[:, 0], reference_stokes_post_circ_polarizer_2)

		# Test power conservation.
		assert np.allclose(test_wf.I, wf_forward_circ_polarizer_1.I + wf_forward_circ_polarizer_2.I)

def test_magnifier():
	pupil_grid = make_pupil_grid(128)
	wf = Wavefront(circular_aperture(1)(pupil_grid))
	wf.total_power = 1

	magnifier = Magnifier(1.0)

	for magnification in [0.2, 3.0, [0.3, 0.3], [0.5, 2]]:
		magnifier.magnification = magnification

		wf_backward = magnifier.backward(wf)
		assert np.abs(wf_backward.total_power - 1) < 1e-12
		assert hash(wf_backward.electric_field.grid) == hash(magnifier.get_input_grid(wf.electric_field.grid, 1))

		wf_forward = magnifier.forward(wf)
		assert np.abs(wf_forward.total_power - 1) < 1e-12
		assert hash(wf_forward.electric_field.grid) == hash(magnifier.get_output_grid(wf.electric_field.grid, 1))

def test_pickle_optical_element():
	import dill as pickle

	fname = 'optical_element.pkl'

	pupil_grid = make_pupil_grid(512)
	focal_grid = make_focal_grid(4, 16)

	elem1 = FraunhoferPropagator(pupil_grid, focal_grid)
	elem2 = SurfaceApodizer(pupil_grid.ones(), lambda wvl: 1.5 + wvl)
	elems = [elem1, elem2]

	for elem in elems:
		wf = Wavefront(pupil_grid.zeros())
		img_saved = elem(wf)

		with open(fname, 'wb') as f:
			pickle.dump(elem, f)

		with open(fname, 'rb') as f:
			elem_loaded = pickle.load(f)
			img_loaded = elem_loaded(wf)

			assert np.allclose(img_saved.electric_field, img_loaded.electric_field)

		os.remove(fname)

def test_step_index_fiber():
	core_radius_multimode = 25e-6 # m
	core_radius_singlemode = 2e-6 # m
	fiber_na = 0.13
	fiber_length = 10 # m
	D_pupil = 1 # m
	wavelength = 1e-6 # m

	multimode_fiber = StepIndexFiber(core_radius_multimode, fiber_na, fiber_length)
	singlemode_fiber = StepIndexFiber(core_radius_singlemode, fiber_na, fiber_length)

	assert np.allclose(multimode_fiber.num_modes(wavelength), 208.4953929730128)
	assert np.allclose(singlemode_fiber.num_modes(wavelength), 1.3343705150272818)

	assert np.allclose(multimode_fiber.V(wavelength), 20.42035224833366)
	assert np.allclose(singlemode_fiber.V(wavelength), 1.6336281798666927)

	assert np.allclose(multimode_fiber.mode_field_radius(wavelength), 1.6688624611223803e-05)
	assert np.allclose(singlemode_fiber.mode_field_radius(wavelength), 3.153705704872073e-06)

	core_radii = [core_radius_singlemode, core_radius_multimode]
	fibers = [singlemode_fiber, multimode_fiber]

	for fiber, core_radius in zip(fibers, core_radii):
		pupil_grid = make_pupil_grid(128)
		focal_grid = make_pupil_grid(128, 4 * core_radius)
		focal_length = D_pupil / (2 * fiber_na)

		prop = FraunhoferPropagator(pupil_grid, focal_grid, focal_length=focal_length)

		wf = Wavefront(circular_aperture(D_pupil)(pupil_grid), wavelength)
		wf.total_power = 1
		img = prop(wf)

		forward_wf = fiber.forward(img)
		backward_wf = fiber.backward(img)

		assert forward_wf.total_power <= img.total_power * (1 + 1e-4)
		assert backward_wf.total_power <= forward_wf.total_power * (1 + 1e-4)



from .atmospheric_model import AtmosphericLayer
from ..util import inverse_tikhonov

class ModalAdaptiveOpticsLayer(AtmosphericLayer):
	def __init__(self, layer, controlled_modes, lag):
		self.layer = layer
		self.controlled_modes = controlled_modes

		AtmosphericLayer.__init__(self, layer.input_grid, layer.Cn_squared, layer.L0, layer.velocity, layer.height)

		self.transformation_matrix = controlled_modes.transformation_matrix
		self.transformation_matrix_inverse = inverse_tikhonov(self.transformation_matrix, 1e-7)

		self.corrected_coeffs = []
		self.lag = lag

	def phase_for(self, wavelength):
		ps = self.layer.phase_for(wavelength)
		ps -= self.transformation_matrix.dot(self.corrected_coeffs[0] / wavelength)

		return ps

	def evolve_until(self, t):
		self.layer.evolve_until(t)

		coeffs = self.transformation_matrix_inverse.dot(self.layer.phase_for(1))
		if len(self.corrected_coeffs) > self.lag:
			self.corrected_coeffs.pop(0)
		self.corrected_coeffs.append(coeffs)

	@property
	def Cn_squared(self):
		return self.layer.Cn_squared

	@Cn_squared.setter
	def Cn_squared(self, Cn_squared):
		self.layer.Cn_squared = Cn_squared

	@property
	def outer_scale(self):
		return self.layer.L0

	@outer_scale.setter
	def L0(self, L0):
		self.layer.L0 = L0

	def reset(self):
		self.corrected_coeffs = []
		self.layer.reset()



from hcipy import *
import numpy as np
from math import *
import mpmath
import scipy
import matplotlib.pyplot as plt

import pytest

def zernike_variance_von_karman(n, m, R, k0, Cn_squared, wavelength):
	'''Calculate the variance of the Zernike mode (`n`,`m`), using a von Karman turbulence spectrum.

	Parameters
	----------
	n : int
		The radial Zernike order.
	m : int
		The azimuthal Zernike order.
	R : scalar
		The radius of the aperture.
	k0 : scalar
		The spatial frequency of the outer scale (1/L0).
	Cn_squared : scalar
		The integrated Cn^2 profile.
	wavelength : scalar
		The wavelength at which to calculate the variance.

	Returns
	-------
	scalar
		The variance of the specific Zernike mode.
	'''
	A = 0.00969 * (2*np.pi / wavelength)**2 * Cn_squared
	coeffs_all = (-1)**(n - m) * 2 * (2 * np.pi)**(11./3) * (n + 1) * A * R**(5./3) / (sqrt(np.pi) * np.sin(np.pi * (n + 1./6)))

	term11 = mpmath.hyper([n + (3./2), n + 2, n + 1],[n + (1./6), n + 2, n + 2, 2 * n + 3], (2*np.pi * R * k0)**2)
	term12 = sqrt(np.pi) * (2*np.pi * R * k0)**(2 * n - 5./3) * scipy.special.gamma(n + 1) / (2**(2 * n + 3) * scipy.special.gamma(11./6) * scipy.special.gamma(n + 1./6) * scipy.special.gamma(n + 2)**2)
	term21 = -1 * scipy.special.gamma(7./3) * scipy.special.gamma(17./6) / (2 * scipy.special.gamma(-n + 11./6) * scipy.special.gamma(17./6)**2 * scipy.special.gamma(n + 23./6))
	term22 = mpmath.hyper([11./6, 7./3, 17./6], [-n + 11./6, 17./6, 17./6, n + 23./6], (2*np.pi * R * k0)**2)

	return coeffs_all * (term11 * term12 + term21 * term22)

def check_total_variance(wavelength, D_tel, fried_parameter, outer_scale, propagate_phase_screen):
	velocity = 10.0 # meters/sec
	num_modes = 1000

	pupil_grid = make_pupil_grid(64, D_tel)
	aperture = circular_aperture(D_tel)(pupil_grid)
	wf = Wavefront(aperture, wavelength)

	Cn_squared = Cn_squared_from_fried_parameter(fried_parameter, wavelength)
	layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared, outer_scale, [velocity / np.sqrt(2), velocity / np.sqrt(2)], use_interpolation=False)

	num_iterations = 2000
	total_variance = []

	for it in range(num_iterations):
		layer.reset()
		if propagate_phase_screen:
			layer.t = np.sqrt(2) * D_tel / velocity

		phase = layer.phase_for(wavelength)
		total_variance.append(np.var(phase[aperture>0]))

	variance_measured = np.mean(total_variance)

	variance_theory = 0
	for i in range(num_modes):
		n, m = noll_to_zernike(i + 2)
		variance_theory += zernike_variance_von_karman(n, m, D_tel / 2., 1. / outer_scale, layer.Cn_squared, wavelength)

	assert (variance_measured / variance_theory - 1) < 0.1

def check_zernike_variances(wavelength, D_tel, fried_parameter, outer_scale, propagate_phase_screen):
	velocity = 10.0 # meters/sec
	num_modes = 50

	pupil_grid = make_pupil_grid(128, D_tel)
	aperture = circular_aperture(D_tel)(pupil_grid)

	Cn_squared = Cn_squared_from_fried_parameter(fried_parameter, wavelength)
	layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared, outer_scale, [velocity / np.sqrt(2), velocity / np.sqrt(2)], use_interpolation=False)

	zernike_modes = make_zernike_basis(num_modes + 20, D_tel, pupil_grid, starting_mode=2, radial_cutoff=False)

	weights = evaluate_supersampled(circular_aperture(D_tel), pupil_grid, 32)
	zernike_modes = ModeBasis([z * np.sqrt(weights) for z in zernike_modes])

	transformation_matrix = zernike_modes.transformation_matrix
	projection_matrix = inverse_tikhonov(transformation_matrix, 1e-9)

	num_iterations = 300
	mode_coeffs = []

	for it in range(num_iterations):
		layer.reset()
		if propagate_phase_screen:
			layer.t = np.sqrt(2) * D_tel / velocity

		phase = layer.phase_for(wavelength)
		coeffs = projection_matrix.dot(phase * np.sqrt(weights))[:num_modes]
		mode_coeffs.append(coeffs)

	variances_simulated = np.var(mode_coeffs, axis=0)

	variances_theory = []
	for j in range(num_modes):
		n, m = noll_to_zernike(j + 2)
		variances_theory.append(zernike_variance_von_karman(n, m, D_tel / 2., 1. / outer_scale, layer.Cn_squared, wavelength))
	variances_theory = np.array(variances_theory)
	'''
	plt.plot(variances_simulated, label='simulated')
	plt.plot(variances_theory, label='theory')
	plt.yscale('log')
	plt.xlabel('Noll index')
	plt.ylabel('Variance (rad^2)')
	plt.legend()
	plt.show()
	'''
	assert np.all(np.abs(variances_simulated / variances_theory - 1) < 1)

def test_finite_atmosphere_total_variance():
	check_total_variance(0.5e-6, 1, 0.1, 10, False)

@pytest.mark.slow
def test_finite_atmosphere_total_variance_in_depth():
	# Check some selected parameter sets.
	check_total_variance(2e-6, 1, 0.1, 10, False)
	check_total_variance(0.5e-6, 8, 0.1, 10, False)
	check_total_variance(0.5e-6, 1, 0.3, 10, False)
	check_total_variance(0.5e-6, 1, 0.1, 40, False)

def test_infinite_atmosphere_total_variance():
	check_total_variance(0.5e-6, 1, 0.1, 10, True)

@pytest.mark.slow
def test_infinite_atmosphere_total_variance_in_depth():
	# Check some selected parameter sets.
	check_total_variance(2e-6, 1, 0.1, 10, True)
	check_total_variance(0.5e-6, 8, 0.1, 10, True)
	check_total_variance(0.5e-6, 1, 0.3, 10, True)
	check_total_variance(0.5e-6, 1, 0.1, 40, True)

def test_finite_atmosphere_zernike_variances():
	check_zernike_variances(0.5e-6, 1, 0.1, 10, False)

@pytest.mark.slow
def test_finite_atmosphere_zernike_variances_in_depth():
	# Check some selected parameter sets.
	check_zernike_variances(2e-6, 1, 0.1, 10, False)
	check_zernike_variances(0.5e-6, 8, 0.1, 10, False)
	check_zernike_variances(0.5e-6, 1, 0.3, 10, False)
	check_zernike_variances(0.5e-6, 1, 0.1, 40, False)

def test_infinite_atmosphere_zernike_variances():
	check_zernike_variances(0.5e-6, 1, 0.1, 10, True)

@pytest.mark.slow
def test_infinite_atmosphere_zernike_variances_in_depth():
	# Check some selected parameter sets.
	check_zernike_variances(2e-6, 1, 0.1, 10, True)
	check_zernike_variances(0.5e-6, 8, 0.1, 10, True)
	check_zernike_variances(0.5e-6, 1, 0.3, 10, True)
	check_zernike_variances(0.5e-6, 1, 0.1, 40, True)

def test_multi_layer_atmosphere():
	r0 = 0.1
	wavelength = 500e-9

	pupil_grid = make_pupil_grid(256, 1.5)
	layers = make_standard_atmospheric_layers(pupil_grid)

	atmospheric_model = MultiLayerAtmosphere(layers, scintillation=False)
	atmospheric_model.Cn_squared = Cn_squared_from_fried_parameter(r0, wavelength)
	atmospheric_model.reset()

	aperture = Field(np.exp(-(pupil_grid.as_('polar').r / 0.65)**20), pupil_grid)
	wf = Wavefront(aperture, wavelength)

	# Model with no scintillation should only not modify amplitude
	wf_out = atmospheric_model.forward(wf)
	assert np.allclose(wf_out.amplitude, aperture)

	# It only should modify phase
	electric_field_compare = aperture * np.exp(1j * atmospheric_model.phase_for(wavelength))
	electric_field_simulated = wf_out.electric_field.copy()

	electric_field_compare /= electric_field_compare.mean()
	electric_field_simulated /= electric_field_simulated.mean()
	assert np.allclose(electric_field_compare, electric_field_simulated)

	wf_in = atmospheric_model.backward(wf_out)
	assert np.allclose(wf_in.electric_field, aperture)

	# Scintillation should also modify amplitude of the wavefront
	atmospheric_model.scintillation = True
	wf_out = atmospheric_model.forward(wf)
	assert not np.allclose(wf_out.amplitude, aperture)

	# Trying to get the phase for a model with scintillation should throw an exception
	with pytest.raises(ValueError):
		phase = atmospheric_model.phase_for(wavelength)

	# Evolving a wavefront should modify the underlying layers
	atmospheric_model.evolve_until(1)
	wf_out2 = atmospheric_model.forward(wf)
	assert not np.allclose(wf_out.electric_field, wf_out2.electric_field)

def test_fried_parameter_seeing():
	for i in range(10):
		seeing = np.random.uniform(0.5, 2) # arcsec
		wavelength = np.random.uniform(500e-9, 800e-9)

		fried_parameter = seeing_to_fried_parameter(seeing, wavelength)
		seeing_recovered = fried_parameter_to_seeing(fried_parameter, wavelength)

		assert np.allclose(seeing_recovered, seeing)

def test_phase_covariance_phase_structure_function_von_karman():
	grid = make_uniform_grid([256, 256], [1, 1], has_center=True)

	for r0 in [0.1, 0.2]:
		for L0 in [10, 20]:
			phase_structure_function = phase_structure_function_von_karman(r0, L0)(grid)
			phase_covariance = phase_covariance_von_karman(r0, L0)(grid)

			phase_structure_function_from_covariance = 2 * (phase_covariance.max() - phase_covariance)

			assert np.allclose(phase_structure_function_from_covariance, phase_structure_function)



from copy import copy
import numpy as np

from ..field import Field

def imshow_field(field, grid=None, ax=None, vmin=None, vmax=None, aspect='equal', norm=None, interpolation=None,
		non_linear_axes=False, cmap=None, mask=None, mask_color='k', grid_units=1, *args, **kwargs):
	'''Display a two-dimensional image on a matplotlib figure.

	This function serves as an easy replacement for the matplotlib.pyplot.imshow() function.
	Its signature mostly folows that of matplotlib, with a few minor differences.

	Parameters
	----------
	field : Field or ndarray
		The field that we want to display. If this is an ndarray,
		then the parameter `grid` needs to be supplied. If the field
		is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
		If the field is a vector field with length 3 or 4, these will be
		interpreted as an RGB or RGBA field.
	grid : Grid or None
		If a grid is supplied, it will be used instead of the grid of `field`.
	ax : matplotlib axes
		The axes which to draw on. If it is not given, the current axes will be used.
	vmin : scalar
		The minimum value on the colorbar. If it is not given, then the minimum value
		of the field will be used.
	vmax : scalar
		The maximum value on the colorbar. If it is not given, then the maximum value
		of the field will be used.
	aspect : ['auto', 'equal', scalar]
		If 'auto', changes the image aspect ratio to match that of the axes.
		If 'equal', changes the axes aspect ratio to match that of the image.
	norm : Normalize
		A Normalize instance is used to scale the input to the (0, 1) range for
		input to the `cmap`. If it is not given, a linear scale will be used.
	interpolation : string
		The interpolation method used. The default is 'nearest'. Supported values
		are {'nearest', 'bilinear'}.
	non_linear_axes : boolean
		If axes are scaled in a non-linear way, for example on a log plot, then imshow_field
		needs to use a more expensive implementation. This parameter is to indicate that this
		algorithm needs to be used.
	cmap : Colormap or None
		The colormap with which to plot the image. It is ignored if a complex
		field or a vector field is supplied.
	mask : field or ndarray
		If part of the image needs to be masked, this mask is overlayed on top of the image.
		This is for example useful when plotting a phase pattern on a certain aperture, which
		has no meaning outside of the aperture. Masks can be partially translucent, and will
		be automatically scaled between (0, 1). Zero means invisible, one means visible.
	mask_color : Color
		The color of the mask, if it is used.
	grid_units : scalar or array_like
		The size of a unit square. The grid will be scaled by the inverse of this number before
		plotting. If this is a scalar, an isotropic scaling will be applied.

	Returns
	-------
	AxesImage
		The produced image.
	'''
	import matplotlib as mpl
	import matplotlib.pyplot as plt
	from matplotlib.image import NonUniformImage

	if ax is None:
		ax = plt.gca()

	ax.set_aspect(aspect)

	# Set/Find the correct grid and scale according to received grid units.
	if grid is None:
		if np.allclose(grid_units, 1):
			grid = field.grid
		else:
			grid = field.grid.scaled(1.0 / grid_units)
	else:
		if np.allclose(grid_units, 1):
			field = Field(field, grid)
		else:
			grid = grid.scaled(1.0 / grid_units)
			field = Field(field, grid)

	# If field is complex, draw complex
	if np.iscomplexobj(field):
		f = complex_field_to_rgb(field, rmin=vmin, rmax=vmax, norm=norm)
		vmin = None
		vmax = None
		norm = None
	else:
		f = field

	# Automatically determine vmin, vmax, norm if not overridden
	if norm is None and not np.iscomplexobj(field):
		if vmin is None:
			vmin = np.nanmin(f)
		if vmax is None:
			vmax = np.nanmax(f)
		norm = mpl.colors.Normalize(vmin, vmax)

	# Get extent
	c_grid = grid.as_('cartesian')
	min_x, min_y, max_x, max_y = c_grid.x.min(), c_grid.y.min(), c_grid.x.max(), c_grid.y.max()

	if grid.is_separated and grid.is_('cartesian'):
		# We can draw this directly
		x, y = grid.coords.separated_coords
		z = f.shaped
		if np.iscomplexobj(field) or field.tensor_order > 0:
			z = np.rollaxis(z, 0, z.ndim)
	else:
		# We can't draw this directly.
		raise NotImplementedError()

	if non_linear_axes:
		# Use pcolormesh to display
		x_mid = (x[1:] + x[:-1]) / 2
		y_mid = (y[1:] + y[:-1]) / 2

		x2 = np.concatenate(([1.5 * x[0] - 0.5 * x[1]], x_mid, [1.5 * x[-1] - 0.5 * x[-2]]))
		y2 = np.concatenate(([1.5 * y[0] - 0.5 * y[1]], y_mid, [1.5 * y[-1] - 0.5 * y[-2]]))
		X, Y = np.meshgrid(x2, y2)

		im = ax.pcolormesh(X, Y, z, *args, norm=norm, rasterized=True, cmap=cmap, **kwargs)
	else:
		# Use NonUniformImage to display
		im = NonUniformImage(ax, extent=(min_x, max_x, min_y, max_y), interpolation=interpolation , norm=norm, cmap=cmap, *args, **kwargs)
		im.set_data(x, y, z)

		from matplotlib.patches import Rectangle
		patch = Rectangle((min_x, min_y), max_x - min_x, max_y - min_y, facecolor='none')
		ax.add_patch(patch)
		im.set_clip_path(patch)

		ax.images.append(im)

	ax.set_xlim(min_x, max_x)
	ax.set_ylim(min_y, max_y)

	if mask is not None:
		one = np.ones(grid.size)
		col = mpl.colors.to_rgb(mask_color)

		m = np.array([one * col[0], one * col[1], one * col[2], 1 - mask / np.nanmax(mask)])

		imshow_field(m, grid, ax=ax)

	num_rows, num_cols = field.grid.shape
	def format_coord(x, y): # pragma: no cover
		col = int(np.round((x - min_x) / (max_x - min_x) * (num_cols - 1)))
		row = int(np.round((y - min_y) / (max_y - min_y) * (num_rows - 1)))

		if col >= 0 and col < num_cols and row >= 0 and row < num_rows:
			z = field.shaped[row, col]
			if np.iscomplexobj(z):
				return 'x=%0.3g, y=%0.3g, z=%0.3g + 1j * %0.3g = %0.3g * exp(1j * %0.2f)' % (x, y, z.real, z.imag, np.abs(z), np.angle(z))
			else:
				return 'x=%0.3g, y=%0.3g, z=%0.3g' % (x, y, z)
		return 'x=%0.3g, y=%0.3g' % (x, y)

	ax.format_coord = format_coord

	ax._sci(im)

	return im

def imsave_field(filename, field, grid=None, vmin=None, vmax=None, norm=None, mask=None, mask_color='k', cmap=None):
	'''Save a two-dimensional field as an image.

	Parameters
	----------
	filename : str
		The path to the file in which to save the field.
	field : Field or ndarray
		The field that we want to display. If this is an ndarray,
		then the parameter `grid` needs to be supplied. If the field
		is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
		If the field is a vector field with length 3 or 4, these will be
		interpreted as an RGB or RGBA field.
	grid : Grid or None
		If a grid is supplied, it will be used instead of the grid of `field`.
	vmin : scalar
		The minimum value on the colorbar. If it is not given, then the minimum value
		of the field will be used.
	vmax : scalar
		The maximum value on the colorbar. If it is not given, then the maximum value
		of the field will be used.
	norm : Normalize
		A Normalize instance is used to scale the input to the (0, 1) range for
		input to the `cmap`. If it is not given, a linear scale will be used.
	mask : field or ndarray
		If part of the image needs to be masked, this mask is overlayed on top of the image.
		This is for example useful when plotting a phase pattern on a certain aperture, which
		has no meaning outside of the aperture. Masks can be partially translucent, and will
		be automatically scaled between (0, 1). Zero means invisible, one means visible.
	mask_color : Color
		The color of the mask, if it is used.
	cmap : Colormap or None
		The colormap with which to plot the image. It is ignored if a complex
		field or a vector field is supplied.
	'''
	import matplotlib as mpl
	import matplotlib.pyplot as plt

	if grid is None:
		grid = field.grid
	else:
		field = Field(field, grid)

	# If field is complex, draw complex
	if np.iscomplexobj(field):
		f = complex_field_to_rgb(field, rmin=vmin, rmax=vmax, norm=norm)
		vmin = None
		vmax = None
		norm = None
	else:
		if norm is None:
			if vmin is None:
				vmin = np.nanmin(field)
			if vmax is None:
				vmax = np.nanmax(field)
			norm = mpl.colors.Normalize(vmin, vmax)
		f = field

	if mask is not None:
		f[~mask.astype('bool')] = np.nan

	cmap = copy(mpl.cm.get_cmap(cmap))
	cmap.set_bad(mask_color)

	plt.imsave(filename, f.shaped, cmap=cmap, vmin=vmin, vmax=vmax)

def contour_field(field, grid=None, ax=None, grid_units=1, *args, **kwargs):
	'''Plot contours of a field.

	Parameters
	----------
	field : Field or ndarray
		The field that we want to display. If this is an ndarray,
		then the parameter `grid` needs to be supplied. If the field
		is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
		If the field is a vector field with length 3 or 4, these will be
		interpreted as an RGB or RGBA field.
	grid : Grid or None
		If a grid is supplied, it will be used instead of the grid of `field`.
	ax : matplotlib axes
		The axes which to draw on. If it is not given, the current axes will be used.
	grid_units : scalar or array_like
		The size of a unit square. The grid will be scaled by the inverse of this number before
		plotting. If this is a scalar, an isotropic scaling will be applied.

	Returns
	-------
	QuadContourSet
		The plotted contour set.
	'''
	import matplotlib.pyplot as plt
	if ax is None:
		ax = plt.gca()

	# Set/Find the correct grid and scale according to received grid units.
	if grid is None:
		if np.allclose(grid_units, 1):
			grid = field.grid
		else:
			grid = field.grid.scaled(1.0 / grid_units)
	else:
		if np.allclose(grid_units, 1):
			field = Field(field, grid)
		else:
			grid = grid.scaled(1.0 / grid_units)
			field = Field(field, grid)

	c_grid = grid.as_('cartesian')
	min_x, min_y, max_x, max_y = c_grid.x.min(), c_grid.y.min(), c_grid.x.max(), c_grid.y.max()

	if grid.is_separated and grid.is_('cartesian'):
		# We can contour directly
		x, y = grid.coords.separated_coords
		z = field.shaped

		X, Y = np.meshgrid(x, y)

		cs = ax.contour(X, Y, z, *args, **kwargs)
	else:
		raise NotImplementedError()

	ax.set_xlim(min_x, max_x)
	ax.set_ylim(min_y, max_y)

	return cs

def contourf_field(field, grid=None, ax=None, grid_units=1, *args, **kwargs):
	'''Plot filled contours of a field.

	Parameters
	----------
	field : Field or ndarray
		The field that we want to display. If this is an ndarray,
		then the parameter `grid` needs to be supplied. If the field
		is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
		If the field is a vector field with length 3 or 4, these will be
		interpreted as an RGB or RGBA field.
	grid : Grid or None
		If a grid is supplied, it will be used instead of the grid of `field`.
	ax : matplotlib axes
		The axes which to draw on. If it is not given, the current axes will be used.
	grid_units : scalar or array_like
		The size of a unit square. The grid will be scaled by the inverse of this number before
		plotting. If this is a scalar, an isotropic scaling will be applied.

	Returns
	-------
	QuadContourSet
		The plotted contour set.
	'''
	import matplotlib.pyplot as plt
	if ax is None:
		ax = plt.gca()

	# Set/Find the correct grid and scale according to received grid units.
	if grid is None:
		if np.allclose(grid_units, 1):
			grid = field.grid
		else:
			grid = field.grid.scaled(1.0 / grid_units)
	else:
		if np.allclose(grid_units, 1):
			field = Field(field, grid)
		else:
			grid = grid.scaled(1.0 / grid_units)
			field = Field(field, grid)

	c_grid = grid.as_('cartesian')
	min_x, min_y, max_x, max_y = c_grid.x.min(), c_grid.y.min(), c_grid.x.max(), c_grid.y.max()

	if grid.is_separated and grid.is_('cartesian'):
		# We can contour directly
		x, y = grid.coords.separated_coords
		z = field.shaped

		X, Y = np.meshgrid(x, y)

		cs = ax.contourf(X, Y, z, *args, **kwargs)
	else:
		raise NotImplementedError()

	ax.set_xlim(min_x, max_x)
	ax.set_ylim(min_y, max_y)

	return cs

def complex_field_to_rgb(field, theme='dark', rmin=None, rmax=None, norm=None):
	'''Convert a complex field to an RGB field.

	This function takes a scalar Field with complex numbers and converts it to
	a vector Field containing the red, green and blue values corresponding to the
	phase and amplitude of the complex input Field. The phase is decoded as hue, and
	amplitude as saturation/value. This function is especially useful for plotting complex
	fields.

	Parameters
	----------
	field : scalar Field or ndarray
		The field that we want to convert.
	theme : ['dark', 'light']
		Whether to modulate saturation or value as a function of amplitude of the
		input field. Dark mode modulates value, light mode modulates saturation.
	rmin : scalar
		The minimum value on the colorbar. If this is not given, then the minimum
		of the absolute value of the Field will be used.
	rmax : scalar
		The maximum value on the colorbar. If this is not given, then the maximum
		of the absolute value of the Field will be used.
	norm : Normalize
		A Normalize instance is used to scale the absolute value to the (0, 1) range
		used for the value or saturation for the returned color. If it is not given,
		a linear scale will be used.

	Returns
	-------
	Field
		A vector field containing the red, green and blue components.

	Raises
	------
	ValueError
		If the supplied field is not a scalar field.
	'''
	import matplotlib as mpl

	if not field.is_scalar_field:
		raise ValueError('Field must be a scalar field.')

	if norm is None:
		if rmin is None:
			rmin = np.nanmin(np.abs(field))
		if rmax is None:
			rmax = np.nanmax(np.abs(field))
		norm = mpl.colors.Normalize(rmin, rmax, True)

	hsv = np.zeros((field.size, 3), dtype='float')
	hsv[..., 0] = np.angle(field) / (2 * np.pi) % 1

	t = norm(np.abs(field))
	if theme == 'light':
		hsv[..., 1] = t
		hsv[..., 2] = 1
	elif theme == 'dark':
		hsv[..., 1] = 1
		hsv[..., 2] = t

	rgb = mpl.colors.hsv_to_rgb(hsv)
	alpha = np.isfinite(field)[:,np.newaxis]

	res = np.concatenate((rgb, alpha), axis=1)

	return Field(res.T, field.grid)



import numpy as np
from ..field import make_hexagonal_grid, Field
from .generic import make_spider, circular_aperture, hexagonal_aperture, make_segmented_aperture, make_spider_infinite, make_obstructed_circular_aperture, rectangular_aperture, make_obstruction

_vlt_telescope_aliases = {'antu': 'ut1', 'kueyen': 'ut2', 'melipal': 'ut3', 'yepun': 'ut4'}

def make_vlt_aperture(normalized=False, telescope='ut3', with_spiders=True, with_M3_cover=False):
	'''Make the VLT aperture.

	This aperture is based on the ERIS pupil documentation: VLT-SPE-AES-11310-0006.

	Parameters
	----------
	normalized : boolean
		If this is True, the outer diameter will be scaled to 1. Otherwise, the
		diameter of the pupil will be 8.0 meters for UT1-3 and 8.1196 meters for UT4.
	telescope : one of {'ut1', 'ut2', 'ut3', 'ut4', 'antu', 'kueyen', 'melipal', 'yepun'}
		The specific telescope on the VLT, case insensitive. Default: UT3.
	with_spiders : boolean
		If this is False, the spiders will be left out. Default: True.
	with_M3_cover : boolean
		If this is True, a cover will be created for the M3 in stowed position.
		This M3 cover is only available on UT4, mimicking the ERIS pupil. A warning
		will be emitted when using an M3 cover with other UTs. Default: False.

	Returns
	-------
	Field generator
		The VLT aperture.
	'''
	telescope = telescope.lower()
	if telescope not in ['ut1', 'ut2', 'ut3', 'ut4']:
		if telescope not in _vlt_telescope_aliases:
			raise ValueError(f'The VLT telescope "{telescope}" has not been implemented.')
		telescope = _vlt_telescope_aliases[telescope]

	if telescope in ['ut1', 'ut2', 'ut3']:
		pupil_diameter = 8.0 # meter
		central_obscuration_ratio = 1.116 / pupil_diameter
	elif telescope == 'ut4':
		pupil_diameter = 8.1196 # meter
		central_obscuration_ratio = 0.6465 * 2 / pupil_diameter

	spider_width = 0.040 # meter
	spider_offset = 0.4045 # meter
	spider_outer_radius = 4.2197 # meter
	outer_diameter_M3_stow = 1.070 # meter
	angle_between_spiders = 101 # degrees

	if with_M3_cover and telescope != 'ut4':
		import warnings
		warnings.warn('Using the M3 cover on a telescope other than UT4 is not realistic.', stacklevel=2)

	if normalized:
		spider_width /= pupil_diameter
		spider_offset /= pupil_diameter
		spider_outer_radius /= pupil_diameter
		outer_diameter_M3_stow /= pupil_diameter
		pupil_diameter = 1.0

	obstructed_aperture = make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio)

	if with_spiders:
		spider_inner_radius = spider_offset / np.cos(np.radians(45 - (angle_between_spiders - 90) / 2))

		spider_start_1 = -spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
		spider_end_1 = spider_outer_radius * np.array([np.cos(np.pi), np.sin(np.pi)])

		spider_start_2 = -spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
		spider_end_2 = spider_outer_radius * np.array([np.cos(-np.pi / 2), np.sin(-np.pi / 2)])

		spider_start_3 = spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
		spider_end_3 = spider_outer_radius * np.array([np.cos(0), np.sin(0)])

		spider_start_4 = spider_inner_radius * np.array([np.cos(np.pi / 4), np.sin(np.pi / 4)])
		spider_end_4 = spider_outer_radius * np.array([np.cos(np.pi / 2), np.sin(np.pi / 2)])

		spider1 = make_spider(spider_start_1, spider_end_1, spider_width)
		spider2 = make_spider(spider_start_2, spider_end_2, spider_width)
		spider3 = make_spider(spider_start_3, spider_end_3, spider_width)
		spider4 = make_spider(spider_start_4, spider_end_4, spider_width)

	if with_M3_cover:
		m3_cover = make_obstruction(rectangular_aperture(outer_diameter_M3_stow, center=[outer_diameter_M3_stow / 2, 0]))

	if with_spiders:
		if with_M3_cover:
			def func(grid):
				return Field(obstructed_aperture(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid) * m3_cover(grid), grid)
		else:
			def func(grid):
				return Field(obstructed_aperture(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid), grid)
	else:
		if with_M3_cover:
			def func(grid):
				return Field(obstructed_aperture(grid) * m3_cover(grid), grid)
		else:
			def func(grid):
				return Field(obstructed_aperture(grid), grid)

	return func

def make_subaru_aperture():
	pass

def make_lbt_aperture():
	pass

def make_magellan_aperture(normalized=False, with_spiders=True):
	'''Make the Magellan aperture.

	Parameters
	----------
	normalized : boolean
		If this is True, the outer diameter will be scaled to 1. Otherwise, the
		diameter of the pupil will be 6.5 meters.
	with_spiders: boolean
		If this is False, the spiders will be left out.

	Returns
	-------
	Field generator
		The Magellan aperture.
	'''
	pupil_diameter = 6.5 # meter
	spider_width1 = 0.75 * 0.0254 # meter
	spider_width2 = 1.5 * 0.0254 # meter
	central_obscuration_ratio = 0.29
	spider_offset = [0, 0.34] # meter

	if normalized:
		spider_width1 /= pupil_diameter
		spider_width2 /= pupil_diameter
		spider_offset = [x / pupil_diameter for x in spider_offset]
		pupil_diameter = 1.0

	spider_offset = np.array(spider_offset)

	mirror_edge1 = (pupil_diameter / (2 * np.sqrt(2)), pupil_diameter / (2 * np.sqrt(2)))
	mirror_edge2 = (-pupil_diameter / (2 * np.sqrt(2)), pupil_diameter / (2 * np.sqrt(2)))
	mirror_edge3 = (pupil_diameter / (2 * np.sqrt(2)), -pupil_diameter / (2 * np.sqrt(2)))
	mirror_edge4 = (-pupil_diameter / (2 * np.sqrt(2)), -pupil_diameter / (2 * np.sqrt(2)))

	obstructed_aperture = make_obstructed_circular_aperture(pupil_diameter, central_obscuration_ratio)

	if not with_spiders:
		return obstructed_aperture

	spider1 = make_spider(spider_offset, mirror_edge1, spider_width1)
	spider2 = make_spider(spider_offset, mirror_edge2, spider_width1)
	spider3 = make_spider(-spider_offset, mirror_edge3, spider_width2)
	spider4 = make_spider(-spider_offset, mirror_edge4, spider_width2)

	def func(grid):
		return obstructed_aperture(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)
	return func

def make_keck_aperture():
	pass

def make_luvoir_a_aperture(normalized=False, with_spiders=True, with_segment_gaps=True,
		gap_padding=1, segment_transmissions=1, return_header=False, return_segments=False):
	'''
	This aperture changes frequently. This one is based on LUVOIR Apertures dimensions
	from Matt Bolcar, LUVOIR lead engineer (as of 10 April 2019)
	Spiders and segment gaps can be included or excluded, and the transmission for each
	of the segments can also be changed. Segements can be returned as well.

	Parameters
	----------
	normalized : boolean
		If this is True, the pupil diameter will be scaled to 1. Otherwise, the
		diameter of the pupil will be 15.0 meters.
	with_spiders : boolean
		Include the secondary mirror support structure in the aperture.
	with_segment_gaps : boolean
		Include the gaps between individual segments in the aperture.
	gap_padding : scalar
		Arbitrary padding of gap size to represent gaps on smaller arrays - this effectively
		makes the gaps larger and the segments smaller to preserve the same segment pitch.
	segment_transmissions : scalar or array_like
		The transmission for each of the segments. If this is a scalar, this transmission
		will be used for all segments.
	return_header : boolean
		If this is True, a header will be returned giving all important values for the
		created aperture for reference.
	return_segments : boolean
		If this is True, the segments will also be returned as a ModeBasis.

	Returns
	-------
	aperture : Field generator
		The LUVOIR A aperture.
	aperture_header : dict
		A dictionary containing all quantities used when making this aperture. Only returned if
		`return_header` is True.
	segments : list of Field generators
		The segments. Only returned when `return_segments` is True.
	'''
	pupil_diameter = 15.0 #m actual circumscribed diameter, used for lam/D calculations other measurements normalized by this diameter
	pupil_inscribed = 13.5 #m actual inscribed diameter
	actual_segment_flat_diameter = 1.2225 #m actual segment flat-to-flat diameter
	actual_segment_gap = 0.006 #m actual gap size between segments
	spider_width = 0.150 #m actual strut size
	lower_spider_angle = 12.7 #deg angle at which lower spiders are offset from vertical
	spid_start = 0.30657 #m spider starting point distance from center of aperture

	segment_gap = actual_segment_gap * gap_padding #padding out the segmentation gaps so they are visible and not sub-pixel
	if not with_segment_gaps:
		segment_gap = 0

	segment_flat_diameter = actual_segment_flat_diameter - (segment_gap - actual_segment_gap)
	segment_circum_diameter = 2 / np.sqrt(3) * segment_flat_diameter #segment circumscribed diameter

	num_rings = 6 #number of full rings of hexagons around central segment

	lower_spider_angle = 12.7 #deg spiders are upside-down 'Y' shaped; degree the lower two spiders are offset from vertical by this amount

	if not with_segment_gaps:
		segment_gap = 0

	aperture_header = {'TELESCOP':'LUVOIR A','D_CIRC': pupil_diameter, 'D_INSC':pupil_inscribed,\
						'SEG_F2F_D':actual_segment_flat_diameter,'SEG_GAP':actual_segment_gap, \
						'STRUT_W':spider_width,'STRUT_AN':lower_spider_angle,'NORM':normalized, \
						'SEG_TRAN':segment_transmissions,'GAP_PAD':gap_padding, 'STRUT_ST':spid_start, \
						'PROV':'MBolcar ppt 20180815'}

	if normalized:
		segment_circum_diameter /= pupil_diameter
		actual_segment_flat_diameter /= pupil_diameter
		actual_segment_gap /= pupil_diameter
		spider_width /= pupil_diameter
		spid_start /= pupil_diameter
		pupil_diameter = 1.0

	segment_positions = make_hexagonal_grid(actual_segment_flat_diameter + actual_segment_gap, num_rings)
	segment_positions = segment_positions.subset(circular_aperture(pupil_diameter * 0.98)) #corner clipping
	segment_positions = segment_positions.subset(lambda grid: ~(circular_aperture(segment_circum_diameter)(grid) > 0))

	segment = hexagonal_aperture(segment_circum_diameter, np.pi / 2)

	if with_spiders:
		spider1 = make_spider_infinite([0, 0], 90, spider_width)
		spider2 = make_spider_infinite([spid_start, 0], 270 - lower_spider_angle, spider_width)
		spider3 = make_spider_infinite([-spid_start, 0], 270 + lower_spider_angle, spider_width)

	segmented_aperture = make_segmented_aperture(segment, segment_positions, segment_transmissions, return_segments=return_segments)

	if return_segments:
		segmented_aperture, segments = segmented_aperture

	def func(grid):
		res = segmented_aperture(grid)

		if with_spiders:
			res *= spider1(grid) * spider2(grid) * spider3(grid)

		return Field(res, grid)

	if with_spiders and return_segments:
		# Use function to return the lambda, to avoid incorrect binding of variables
		def segment_with_spider(segment):
			return lambda grid: segment(grid) * spider1(grid) * spider2(grid) * spider3(grid)

		segments = [segment_with_spider(s) for s in segments]

	if return_header:
		if return_segments:
			return func, aperture_header, segments
		else:
			return func, aperture_header
	elif return_segments:
		return func, segments
	else:
		return func

def make_luvoir_a_lyot_stop(normalized=False, with_spiders=False, spider_oversize=1,
		inner_diameter_fraction=0.2, outer_diameter_fraction=0.9, return_header=False):
	'''Make a LUVOIR-A Lyot stop for the APLC coronagraph.

	Parameters
	----------
	normalized : boolean
		If this is True, the pupil diameter will be scaled to 1. Otherwise, the
		diameter of the pupil will be 15.0 meters.
	with_spiders : boolean
		Include the secondary mirror support structure in the aperture.
	inner_diameter_fraction : scalar
		The fractional size of the circular central obstruction as fraction of the pupil diameter.
	outer_diameter_fraction : scalar
		The fractional size of the circular outer edge as fraction of the pupil diameter.
	spider_oversize : scalar
		The factor by which to oversize the spiders compared to the LUVOIR-A aperture spiders.
	return_header : boolean
		If this is True, a header will be returned giving all important values for the
		created aperture for reference.

	Returns
	-------
	lyot_stop : Field generator
		A field generator for the Lyot stop.
	header : dict
		A dictionary containing all important values for the created aperture. Only returned
		if `return_header` is True.
	'''
	pupil_diameter = 15.0 #m actual circumscribed diameter, used for lam/D calculations other measurements normalized by this diameter
	spider_width = 0.150 #m actual strut size
	lower_spider_angle = 12.7 #deg angle at which lower spiders are offset from vertical
	spid_start = 0.30657 #m spider starting point offset from center of aperture

	outer_D = pupil_diameter * outer_diameter_fraction
	inner_D = pupil_diameter * inner_diameter_fraction
	pad_spid_width = spider_width * spider_oversize

	lyot_reference_diameter = pupil_diameter

	ls_header = {'TELESCOP':'LUVOIR A', 'D_CIRC': pupil_diameter, 'D_INSC': 13.5,
				 'LS_ID': inner_diameter_fraction, 'LS_OD': outer_diameter_fraction, 'LS_REF_D': lyot_reference_diameter, 'NORM': normalized, 'STRUT_ST': spid_start}

	if with_spiders:
		ls_header['STRUT_W']  = spider_width
		ls_header['STRUT_AN'] = lower_spider_angle
		ls_header['STRUT_P']  = spider_oversize

	if normalized:
		outer_D /= pupil_diameter
		inner_D /= pupil_diameter
		pad_spid_width /= pupil_diameter
		spid_start /= pupil_diameter

	outer_diameter = circular_aperture(outer_D)
	central_obscuration = circular_aperture(inner_D)

	if with_spiders:
		spider1 = make_spider_infinite([0, 0], 90, pad_spid_width)
		spider2 = make_spider_infinite([spid_start,0], 270 - lower_spider_angle, pad_spid_width)
		spider3 = make_spider_infinite([-spid_start,0], 270 + lower_spider_angle, pad_spid_width)

	def aper(grid):
		result = outer_diameter(grid) - central_obscuration(grid)

		if with_spiders:
			result *= spider1(grid) * spider2(grid) * spider3(grid)

		return result

	if return_header:
		return aper, ls_header

	return aper

def make_hicat_aperture(normalized=False, with_spiders=True, with_segment_gaps=True, return_header=False, return_segments=False):
	'''Make the HiCAT P3 apodizer mask

	Parameters
	----------
	normalized : boolean
		If this is True, the outer diameter will be scaled to 1. Otherwise, the
		diameter of the pupil will be 15.0 meters.
	with_spiders : boolean
		Include the secondary mirror support structure in the aperture.
	with_segment_gaps : boolean
		Include the gaps between individual segments in the aperture.
	return_header : boolean
		If this is True, a header will be returned giving all important values for the
		created aperture for reference.
	return_segments : boolean
		If this is True, the segments will also be returned as a list of Field generators.

	Returns
	-------
	aperture : Field generator
		The HiCAT aperture.
	header : dict
		A dictionary containing all important values for the created aperture. Only returned
		if `return_header` is True.
	segments : list of Field generators
		The segments. Only returned when `return_segments` is True.
	'''
	gamma_21 = 0.423
	gamma_31 = 1.008

	#P2 - Iris AO
	p2_irisao_segment_size = 1.4e-3 # m (note: point to point)
	p2_irisao_segment_side_length = p2_irisao_segment_size / 2
	p2_irisao_segment_gap_size = 12e-6 # m

	p2_irisao_distance_between_segments = p2_irisao_segment_side_length * np.sqrt(3)
	p2_irisao_segment_circumdiameter = (2 * p2_irisao_segment_side_length) - (2/np.sqrt(3)) * p2_irisao_segment_gap_size

	#P1 - Pupil Mask
	# Central segment
	p1_pupil_mask_central_segment_size = 3.600e-3 # m

	#P3 - Apodizer
	# Contour
	p3_apodizer_size = 19.725e-3 # m
	p2_apodizer_size = p3_apodizer_size * gamma_21 / gamma_31

	# Gap
	p3_apodizer_mask_gap_size = 0.090e-3 # m
	p3_irisao_segment_gap_size = p2_irisao_segment_gap_size * gamma_31 / gamma_21
	apodizer_mask_gap_oversize_factor_wrt_irisao = p3_apodizer_mask_gap_size / p3_irisao_segment_gap_size

	# Central segment
	p3_apodizer_mask_central_segment_size = 3.950e-3 # m
	p3_pupil_mask_central_segment_size = p1_pupil_mask_central_segment_size * gamma_31
	apodizer_mask_central_segment_oversize_factor_wrt_pupil_mask = p3_apodizer_mask_central_segment_size / p3_pupil_mask_central_segment_size
	p3_irisao_segment_size = p2_irisao_segment_size * gamma_31 / gamma_21

	# Spiders
	p3_apodizer_mask_spiders_thickness = 0.350e-3 # m

	header = {'TELESCOP':'HiCAT', 'P3_APOD': p3_apodizer_size, 'P3_CENT_SEG': p3_apodizer_mask_central_segment_size,
				'P3_GAP': p3_apodizer_mask_gap_size, 'P3_GAP_OVER': apodizer_mask_gap_oversize_factor_wrt_irisao,
				'P3_STRUT': p3_apodizer_mask_spiders_thickness, 'PROV': 'HiCAT spreadsheet'}

	p3_irisao_segment_circumdiameter = p2_irisao_segment_circumdiameter * gamma_31 / gamma_21
	p3_irisao_distance_between_segments = p2_irisao_distance_between_segments * gamma_31 / gamma_21
	p3_apodizer_segment_circumdiameter = p3_irisao_segment_circumdiameter + (-p3_apodizer_mask_gap_size + p3_irisao_segment_gap_size) * (2/np.sqrt(3))

	if normalized:
		p3_apodizer_segment_circumdiameter /= p3_apodizer_size
		p3_irisao_distance_between_segments /= p3_apodizer_size
		p3_apodizer_mask_central_segment_size /= p3_apodizer_size
		p3_apodizer_mask_spiders_thickness /= p3_apodizer_size
		p3_apodizer_size = 1

	segment = hexagonal_aperture(p3_apodizer_segment_circumdiameter, np.pi / 2)
	segment_positions = make_hexagonal_grid(p3_irisao_distance_between_segments, 3, False)
	segmentation = make_segmented_aperture(segment, segment_positions, return_segments=return_segments)

	if return_segments:
		segmentation, segments = segmentation

	segment = hexagonal_aperture(p3_apodizer_size / 7 / np.sqrt(3) * 2, np.pi / 2)
	distance_between_segments = p3_apodizer_size / 7
	segment_positions = make_hexagonal_grid(distance_between_segments, 3)
	contour = make_segmented_aperture(segment, segment_positions)

	central_segment = hexagonal_aperture(p3_apodizer_mask_central_segment_size, np.pi / 2)

	if return_segments:
		# Use function to return the lambda, to avoid incorrect binding of variables
		def segment_obstructed(segment):
			return lambda grid: segment(grid) * (contour(grid) - central_segment(grid))

		segments = [segment_obstructed(s) for s in segments]

	if with_spiders:
		spider1 = make_spider_infinite([0,0], 60, p3_apodizer_mask_spiders_thickness)
		spider2 = make_spider_infinite([0,0], 120, p3_apodizer_mask_spiders_thickness)
		spider3 = make_spider_infinite([0,0], -60, p3_apodizer_mask_spiders_thickness)
		spider4 = make_spider_infinite([0,0], -120, p3_apodizer_mask_spiders_thickness)

	if with_spiders and return_segments:
		# Use function to return the lambda, to avoid incorrect binding of variables
		def segment_with_spider(segment):
			return lambda grid: segment(grid) * spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)

		segments = [segment_with_spider(s) for s in segments]

	def func(grid):
		res = contour(grid) - central_segment(grid)

		if with_segment_gaps:
			res *= segmentation(grid)

		if with_spiders:
			res *= spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)

		return Field(res, grid)

	if return_header:
		if return_segments:
			return func, header, segments
		else:
			return func, header
	elif return_segments:
		return func, segments
	else:
		return func

def make_hicat_lyot_stop(normalized=False, with_spiders=True, inner_diameter_fraction=0.2, outer_diameter_fraction=0.9, return_header=False):
	'''Make the HiCAT Lyot stop.

	Parameters
	----------
	normalized : boolean
		If this is True, the outer diameter will be scaled to 1. Otherwise, the
		diameter of the pupil will be 15.0 meters.
	with_spiders : boolean
		Include the secondary mirror support structure in the aperture.
	inner_diameter_fraction : scalar
		The fractional size of the circular central obstruction as fraction of the pupil diameter.
	outer_diameter_fraction : scalar
		The fractional size of the circular outer edge as fraction of the pupil diameter.
	return_header : boolean
		If this is True, a header will be returned giving all important values for the
		created aperture for reference.

	Returns
	-------
	lyot_stop : Field generator
		A field generator for the Lyot stop.
	header : dict
		A dictionary containing all important values for the created aperture. Only returned
		if `return_header` is True.
	'''
	gamma_21 = 0.423
	gamma_31 = 1.008
	gamma_51 = 0.979

	p2_irisao_segment_size = 1.4e-3 # m (note: point to point)
	p2_irisao_segment_side_length = p2_irisao_segment_size / 2
	p2_irisao_inscribed_circle_size = 10 * p2_irisao_segment_side_length
	p2_irisao_flat_to_flat_size = 14 * np.sqrt(3) / 2 * p2_irisao_segment_side_length
	p2_irisao_circumscribed_circle_size = np.sqrt(p2_irisao_flat_to_flat_size**2 + p2_irisao_segment_side_length**2)

	p3_apodizer_mask_central_segment_size = 3.950e-3 # m
	p3_apodizer_size = 19.725e-3 # m
	p5_apodizer_size = p3_apodizer_size * gamma_51 / gamma_31

	p5_lyot_stop_size = outer_diameter_fraction * p5_apodizer_size # m
	p5_irisao_inscribed_circle_size = p2_irisao_inscribed_circle_size * gamma_51 / gamma_21
	lyot_stop_mask_undersize_contour_wrt_inscribed_circle = p5_lyot_stop_size / p5_irisao_inscribed_circle_size

	p5_irisao_flat_to_flat_size = p2_irisao_flat_to_flat_size * gamma_51 / gamma_21
	p5_irisao_circumscribed_circle_size = p2_irisao_circumscribed_circle_size * gamma_51 / gamma_21

	# Central segment
	p5_lyot_stop_mask_central_segment_size = inner_diameter_fraction * p5_apodizer_size # m
	p5_apodizer_mask_central_segment_size = p3_apodizer_mask_central_segment_size * gamma_51 / gamma_31

	p5_irisao_segment_size = p2_irisao_segment_size * gamma_51 / gamma_21
	lyot_stop_mask_central_segment_oversize_factor_wrt_apodizer_mask = p5_lyot_stop_mask_central_segment_size / p5_apodizer_mask_central_segment_size

	# Spiders
	p5_lyot_stop_mask_spiders_thickness = 0.700e-3 # m
	lyot_stop_mask_spiders_thickness_ratio = p5_lyot_stop_mask_spiders_thickness / p5_irisao_circumscribed_circle_size

	if normalized:
		p5_lyot_stop_size /= p5_apodizer_size
		p5_lyot_stop_mask_central_segment_size /= p5_apodizer_size
		p5_lyot_stop_mask_spiders_thickness /= p5_apodizer_size

	central_obscuration = circular_aperture(p5_lyot_stop_mask_central_segment_size)
	outer_diameter = circular_aperture(p5_lyot_stop_size)

	header = {'TELESCOP':'HiCAT', 'P3_APOD': p3_apodizer_size, 'P3_CENT_SEG': p3_apodizer_mask_central_segment_size,
				'LS_CENT': p5_lyot_stop_mask_central_segment_size, 'LS_SIZE': p5_lyot_stop_size,
				'P5_STRUT': p5_lyot_stop_mask_spiders_thickness}

	if with_spiders:
		spider1 = make_spider_infinite([0,0], 60, p5_lyot_stop_mask_spiders_thickness)
		spider2 = make_spider_infinite([0,0], 120, p5_lyot_stop_mask_spiders_thickness)
		spider3 = make_spider_infinite([0,0], -60, p5_lyot_stop_mask_spiders_thickness)
		spider4 = make_spider_infinite([0,0], -120, p5_lyot_stop_mask_spiders_thickness)

	def func(grid):
		res = outer_diameter(grid) - central_obscuration(grid)

		if with_spiders:
			res *= spider1(grid) * spider2(grid) * spider3(grid) * spider4(grid)

		return Field(res, grid)

	if return_header:
		return func, header
	else:
		return func

def make_elt_aperture():
	pass



import os
import yaml
import pkg_resources

class _ConfigurationItem(object):
	def __init__(self, val):
		self._val = val

	def __getitem__(self, key):
		val = self._val[key]

		if isinstance(val, dict):
			return _ConfigurationItem(val)
		else:
			return val

	def __setitem__(self, key, value):
		self._val[key] = value

	def __contains__(self, key):
		return key in self._val

	def __getattr__(self, name):
		if name == '_val':
			return getattr(self, name)
		else:
			return self.__getitem__(name)

	def __setattr__(self, name, value):
		if name == '_val':
			super().__setattr__(name, value)
		else:
			self.__setitem__(name, value)

	def __str__(self):
		return str(self._val)

	def clear(self):
		self._val.clear()

	def update(self, b):
		for key in b.keys():
			if key in self:
				if isinstance(b[key], dict) and isinstance(self._val[key], dict):
					self[key].update(b[key])
				else:
					self[key] = b[key]
			else:
				self[key] = b[key]

class Configuration(object):
	'''A configuration object that describes the current configuration status of the package.
	'''
	def __init__(self):
		if Configuration._config is None:
			self.reset()

	_config = None

	def __getitem__(self, key):
		'''Get the value for the key.

		Parameters
		----------
		key : string
			The configuration key.
		'''
		return Configuration._config[key]

	def __setitem__(self, key, value):
		'''Set the value for the key.

		Parameters
		----------
		key : string
			The configuration key.
		value : anything
			The value to set this to.
		'''
		Configuration._config[key] = value

	__getattr__ = __getitem__
	__setattr__ = __setitem__

	def __str__(self):
		return str(Configuration._config)

	def reset(self):
		'''Reset the configuration to the default configuration.

		This default configuration consists of the default parameters in `hcipy/data/default_config.yaml`, which
		can be overridden by a configuration file in `~/.hcipy/hcipy_config.yaml`. This can in turn be overridden
		by a configuration file named `hcipy_config.yaml` located in the current working directory.
		'''
		Configuration._config = _ConfigurationItem({})

		default_config = pkg_resources.resource_stream('hcipy', 'data/default_config.yaml')
		user_config = os.path.expanduser('~/.hcipy/hcipy_config.yaml')
		current_working_directory = './hcipy_config.yaml'

		paths = [default_config, user_config, current_working_directory]

		for path in paths:
			try:
				contents = path.read()
			except AttributeError:
				try:
					with open(path) as f:
						contents = f.read()
				except IOError:
					continue

			new_config = yaml.safe_load(contents)
			self.update(new_config)

	def update(self, b):
		'''Update the configuration with the configuration `b`, as a dictionary.

		Parameters
		----------
		b : dict
			A dictionary containing the values to update in the configuration.
		'''
		Configuration._config.update(b)



import numpy as np
from ..field import Field

from ..optics import GeometricPhaseElement

def generate_app_keller(wavefront, propagator, contrast, num_iterations, beta=0):
	"""
	Accelerated Gerchberg-Saxton-like algorithm for APP design by
	Christoph Keller [Keller2016]_ and based on Douglas-Rachford operator splitting.
	The acceleration was inspired by the paper by Jim Fienup [Fienup1976]_. The
	acceleration can provide speed-ups of up to two orders of magnitude and
	produce better APPs.

	.. [Keller2016] Keller C.U., 2016, "Novel instrument concepts for
		characterizing directly imaged exoplanets", Proc. SPIE 9908,
		Ground-based and Airborne Instrumentation for Astronomy VI, 99089V
		doi: 10.1117/12.2232633; https://doi.org/10.1117/12.2232633

	.. [Fienup1976] J. R. Fienup, 1976, "Reconstruction of an object from the modulus
		of its Fourier transform," Opt. Lett. 3, 27-29

	Parameters
	----------
	wavefont : Wavefront
		The input aperture as a wavefront; a phase can be provided as a
		starting point for the APP generation.
	propagator : Propagator
		The propagator from the pupil grid to the focal plane grid.
	contrast : array_like
		The required contrast in the focal plane; this is a float mask
		that is 1.0 everywhere except for the dark zone where it is the
		contrast such as 1e-5.
	num_iterations : int
		The maximum number of iterations.
	beta : scalar
		The acceleration parameter. The default is 0 (no acceleration).
		Good values for beta are typically between 0.3 and 0.9. Values larger
		than 1.0 will not work.

	Returns
	-------
	Wavefront
		The APP as a wavefront.

	Raises
	------
	ValueError
		If beta is not between 0 and 1.
	"""
	if beta < 0 or beta > 1:
		raise ValueError('Beta should be between 0 and 1.')

	# initialize APP with wavefront
	app = wavefront.copy()

	# define pupil aperture where amplitude is larger than 1e-6
	aperture = wavefront.amplitude > 1e-6

	# define dark zone as location where contrast requirement is < 1e-1
	dark_zone = contrast < 0.1

	old_image = None
	for i in range(num_iterations):
		# calculate image plane electric field
		image = propagator.forward(app)

		# stop iteration if contrast requirement is met
		if not np.any(image.intensity / image.intensity.max() > contrast):
			break

		# modify focal plane electic field using acceleration
		new_image = image.copy()
		if beta != 0 and old_image is not None:
			new_image.electric_field[dark_zone] = old_image.electric_field[dark_zone] * beta - new_image.electric_field[dark_zone] * (1 + beta)
		else:
			new_image.electric_field[dark_zone] = 0
		old_image = new_image.copy()

		# determine corresponding aperture electric field
		app.electric_field = propagator.backward(new_image).electric_field

		# enforce aperture support
		app.electric_field[~aperture] = 0

		# enforce unity transmission within aperture support
		app.electric_field[aperture] *= wavefront.amplitude[aperture] / app.amplitude[aperture]

	return app

def generate_app_por(wavefront, propagator, propagator_max, contrast, num_iterations=1):
	'''Optimize a one-sided APP using a globally optimal algorithm. This algorithm does not
	apply any symmetries for two-sided dark zones or circularly-symmetric pupils. This
	function requires that you have installed the Gurobi optimizer.

	Parameters
	----------
	wavefont : Wavefront
		The input aperture as a wavefront.
	propagator : Propagator
		The propagator from `wavefront` to the focal plane grid.
	propagator_max : Propagator
		The propagator from `wavefront` to the point in the focal-plane that we want to maximize.
	contrast : array_like or scalar
		The required contrast in the focal plane.
	num_iterations : int
		The number of iterations for the Strehl ratio to converge.

	Returns
	-------
	Field
		The resultant electric field transmission function.
	'''
	import gurobipy as gp

	M = propagator.get_transformation_matrix_forward(wavefront.electric_field.grid, wavefront.wavelength)
	M_max = propagator_max.get_transformation_matrix_forward(wavefront.electric_field.grid, wavefront.wavelength)

	wf_0 = propagator.forward(wavefront).electric_field
	M /= wf_0[:,np.newaxis]

	M *= wavefront.electric_field
	M_max *= wavefront.electric_field

	M_max = np.sum(M_max, axis=0)

	M = np.array(np.bmat([[M.real, -M.imag], [M.imag, M.real]]))
	M_max = np.concatenate((M_max.real, -M_max.imag))

	m, n = M.shape
	contrast_requirement = np.ones(m) * np.sqrt(contrast)

	model = gp.Model('lp')
	model.Params.Threads = 4

	x = model.addVars(n, lb=-1, ub=1)

	obj = gp.quicksum((x[i] * M_max[i] for i in range(n)))
	model.setObjective(obj, gp.GRB.MAXIMIZE)

	for i in range(n//2):
		r2 = x[i]*x[i] + x[i+n//2]*x[i+n//2]
		model.addQConstr(r2 <= 1)

	for i, ee in enumerate(M):
		e = gp.quicksum((x[i] * ee[i] for i in range(n)))
		model.addConstr(e <= contrast_requirement[i])
		model.addConstr(e >= -contrast_requirement[i])

	model.optimize()

	solution = np.array([x[i].x for i in range(n)])
	solution = solution[:n//2] + 1j * solution[n//2:]

	return Field(solution, wavefront.electric_field.grid)

def generate_app_doelman(wavefront, propagator, propagator_max, contrast, num_iterations):
	raise NotImplementedError()

class VectorApodizingPhasePlate(GeometricPhaseElement):
	'''The vector-Apodizing Phase Plate (vAPP).

	Parameters
	----------
	phase_pattern : field or array_like
		The phase pattern generating the vAPP PSFs (radians).
	leakage : scalar
		The relative leakage strength (0 = no leakage, 1 = maximum leakage)
	retardance_offset : scalar
		The retardance offset from half wave in radians. This will result in leakage.
	'''
	def __init__(self, phase_pattern, leakage=None, retardance_offset=0):

		GeometricPhaseElement.__init__(self, phase_pattern, leakage, retardance_offset)



from .field import imshow_field
from ..field import Field
from ..config import Configuration

import numpy as np

def imshow_psf(psf, grid=None, vmin=1e-8, vmax=1e-1, scale='log',
		cmap=None, title=None, normalization='none',
		crosshairs=False, mark_centroid=False, colorbar=True,
		colorbar_orientation='vertical', spatial_resolution=1,
		ax=None, ticks=None, **kwargs):
	'''Display a PSF in a nice, consistent format.

	Parameters
	----------
	psf : Field or Wavefront
		The point spread function that we want to display.
	grid : Grid or None
		If we want to redefine the grid of the psf.
	vmin : scalar
		The minimum value on the colorbar.
	vmax : scalar
		The maximum value on the colorbar.
	scale : ['linear', 'log', 'logarithmic']
		The scale of the image.
	cmap : matplotlib.cm.Colormap instance
		The colormap to use. If default, it will be taken from the configuration file.
	title : string, optional
		Give the image a title. If this is not given, no title will be displayed.
	normalization : ['peak', 'total', 'none'] or scalar
		The type of normalization to apply. 'peak' normalizes by the maximum intensity.
		'total' normalizes by the total intensity, 'none' doesn't do any normalization.
		If this is a scalar, the PSF will be divided by this number before displaying it.
	crosshairs : boolean
		Whether to draw crosshairs at the image center. Default: False.
	mark_centroid : boolean
		Whether to draw a small crosshair at the centroid of the PSF. The centroid will be
		computed with TODO Default: False.
	colorbar : boolean
		Whether to draw a colorbar on the image. Default: True.
	colorbar_orientation : ['horizontal', 'vertical']
		How the colorbar should be oriented.
	spatial_resolution : scalar
		This will be the size of a unit cell on the axes.
	ax : matplotlib axes
		The axes which to draw on. If it is not given, the current axes will be used.
	ticks : array_like
		The ticks on the added colorbar.
	kwargs : dictionary
		Any keyword arguments will be sent to `imshow_field()`.

	Returns
	-------
	Image
		The created image.
	'''
	import matplotlib as mpl
	import matplotlib.pyplot as plt

	if ax is None:
		ax = plt.gca()

	if hasattr(psf, 'wavelength'):
		psf = psf.power

	if grid is not None:
		psf = Field(psf, grid)

	if normalization.lower() == 'peak':
		psf_norm = psf.max()
	elif normalization.lower() == 'total':
		psf_norm = np.sum(psf * psf.grid.weights)
	elif normalization.lower() == 'none':
		psf_norm = 1
	else:
		psf_norm = normalization

	if cmap is None:
		cmap = Configuration().plotting.psf_colormap

	img = psf / psf_norm

	if scale.lower() == 'linear':
		norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
	elif scale.lower() == 'log' or scale.lower() == 'logarithmic':
		norm = mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
	else:
		raise ValueError('Scale needs to be one of ["linear", "log", "logarithmic"].')

	im = imshow_field(img, norm=norm, cmap=cmap,
		grid_units=spatial_resolution, ax=ax, **kwargs)

	if crosshairs:
		ax.axhline(0, ls=':', color='k')
		ax.axvline(0, ls=':', color='k')

	if title:
		ax.set_title(title)

	if colorbar:
		cb = plt.colorbar(im, ax=ax, orientation=colorbar_orientation)

		if scale.lower() in ['log', 'logarithmic']:
			if ticks is None:
				ticks = np.logspace(np.log10(vmin), np.log10(vmax), int(np.round(np.log10(vmax / vmin) + 1)))

				if colorbar_orientation == 'horizontal' and vmax == 1e-1 and vmin == 1e-8:
					# Use better looking ticks
					ticks = [1e-8, 1e-6, 1e-4, 1e-2, 1e-1]

			cb.set_ticks(ticks)
			cb.set_ticklabels(ticks)

		if normalization.lower() == 'peak':
			cb.set_label('Intensity relative to peak')
		elif normalization.lower() == 'total':
			cb.set_label('Fractional intensity per pixel')

	if mark_centroid:
		raise NotImplementedError()

	return im

def imshow_pupil_phase(pupil_phase, grid=None, phase_limits=None, vmin=None, vmax=None, cmap=None,
		colorbar=True, colorbar_orientation='vertical', title=None, crosshairs=False,
		remove_piston=False, ax=None, **kwargs):
	'''Display a pupil phase pattern in a nice, consistent format.

	Parameters
	----------
	pupil_phase : Field or Wavefront
		The pupil that we want to display. If it is a wavefront or complex, the phase will be used
		as the phase, and the amplitude will be applied as mask.
	grid : Grid or None
		If we want to redefine the grid of the pupil.
	phase_limits : scalar
		The minimum and maximum of the pupil phase. If it is not given, and vmin and vmax are
		not given either, it will be estimated from the data. If the range is sufficiently close
		to 2 pi, the range will be extended to 2 pi. The vmin and vmax, unless overridden, will
		be set to -phase_limits and phase_limits, so the phase will be centered around zero.
	vmin : scalar, optional
		The minimum of the colorbar.
	vmax : scalar, optional
		The maximum of the colorbar.
	cmap : matplotlib.cm.Colormap instance
		The colormap to use. If default, it will be taken from the configuration file.
	colorbar : boolean
		Whether to draw a colorbar on the image. Default: True.
	colorbar_orientation : ['horizontal', 'vertical']
		How the colorbar should be oriented.
	title : string, optional
		Give the image a title. If this is not given, no title will be displayed.
	crosshairs : boolean
		Whether to draw crosshairs at the image center. Default: False.
	remove_piston : boolean
		Whether to remove piston from the phase. Default: False.
	ax : matplotlib axes
		The axes which to draw on. If it is not given, the current axes will be used.
	kwargs : dictionary
		Any keyword arguments will be sent to `imshow_field()`.
	'''
	import matplotlib.pyplot as plt

	if ax is None:
		ax = plt.gca()

	if hasattr(pupil_phase, 'wavelength'):
		pupil_phase = pupil_phase.electric_field

	if grid is not None:
		pupil_phase = Field(pupil_phase, grid)

	if np.iscomplexobj(pupil_phase):
		phase = Field(np.angle(pupil_phase), pupil_phase.grid)
		mask = np.abs(pupil_phase)
	else:
		phase = pupil_phase
		mask = None

	if remove_piston:
		complex_field = np.exp(1j * phase)
		if mask is not None:
			complex_field *= mask

		complex_field /= np.mean(complex_field)
		phase = Field(np.angle(complex_field), phase.grid)

	if phase_limits is None:
		phase_limits = max(-phase[mask > 0].min(), phase[mask > 0].max())

		if np.abs(phase_limits - np.pi) < 0.5:
			phase_limits = np.pi

		if phase_limits < 1e-10:
			phase_limits = 1e-10

	if vmin is None and vmax is None:
		vmin = -phase_limits
		vmax = phase_limits

	if cmap is None:
		cmap = Configuration().plotting.pupil_phase_colormap

	im = imshow_field(phase, ax=ax, cmap=cmap, mask=mask, vmin=vmin, vmax=vmax, **kwargs)

	if crosshairs:
		ax.axhline(0, ls=':', color='k')
		ax.axvline(0, ls=':', color='k')

	if title:
		ax.set_title(title)

	if colorbar:
		cb = plt.colorbar(im, ax=ax, orientation=colorbar_orientation)
		cb.set_label('Phase [rad]')

	return im



import glob
import os
import shutil
import base64
from subprocess import Popen, PIPE

import matplotlib
import imageio
from PIL import Image

class GifWriter(object):
	'''A writer of gif files from Matplotlib figures.

	This function writes out individual frames to a `filename.frames` directory. When
	the `close()` function is called, or the object is removed by the garbage collector,
	the frames are collected in a single gif file. The individual frames are then deleted,
	if `cleanup` is True (default).

	Parameters
	----------
	framerate : integer
		The number of frames per second of the generated gif file.
	cleanup : boolean
		Whether to clean up the generated frames.
	'''
	def __init__(self, filename, framerate=15, cleanup=True):
		self.closed = False
		self.filename = filename
		self.framerate = framerate
		self.cleanup = cleanup

		self.path_to_frames = self.filename + "_frames"
		if not os.path.exists(self.path_to_frames):
			os.mkdir(self.path_to_frames)
		self.num_frames = 0

	def __del__(self):
		try:
			self.close()
		except Exception:
			import warnings
			warnings.warn('Something went wrong while closing the GifWriter...', RuntimeWarning)

	def add_frame(self, fig=None, data=None, cmap=None, dpi=None):
		'''Add a frame to the animation.

		Parameters
		----------
		fig : Matplotlib figure
			The Matplotlib figure acting as the animation frame.
		data : ndarray
			The image data array acting as the animation frame.
		cmap : Matplotlib colormap
			The optional colormap for the image data.
		dpi : integer or None
			The number of dots per inch with which to save the matplotlib figure.
			If it is not given, the default Matplotlib dpi will be used.

		Raises
		------
		RuntimeError
			If the function was called on a closed GifWriter.
		'''
		if self.closed:
			raise RuntimeError('Attempted to add a frame to a closed GifWriter.')

		dest = os.path.join(self.path_to_frames, '%05d.png' % self.num_frames)

		if data is None:
			if fig is None:
				fig = matplotlib.pyplot.gcf()
			fig.savefig(dest, format='png', transparent=False)
		else:
			if cmap is not None:
				data = matplotlib.cm.get_cmap(cmap)(data, bytes=True)

			imageio.imwrite(dest, data, format='png')

		self.num_frames += 1

	@staticmethod
	def convert_to_gif(dest_filename, src_file_path, framerate, src_file_suffix="png", num_files_to_convert=None):
		'''Helper function to convert all files in a directory to a gif file.

		Parameters
		----------
		dest_filename : string
			The filename for the gif file.
		src_file_path : string
			The path to the directory with all the frames as image files.
		framerate : integer
			The number of frames per second for the generated gif file.
		src_file_suffix : string
			The file extension of the image files.
		num_files_to_convert : integer or None
			How many frames are expected in the directory. If None, then no check will be done.

		Raises
		------
		OSError
			If the number of files in the directory differs from the number of expected files.
		'''
		search_pattern = os.path.join(src_file_path, "*."+src_file_suffix)
		files = glob.glob(search_pattern)
		files.sort()

		if num_files_to_convert is not None and len(files) != num_files_to_convert:
			raise OSError("Expected {} files but found {}".format(num_files_to_convert, len(files)))

		# Open all frames to convert
		frames = []
		for image_file in files:
			frames.append(Image.open(image_file).copy())

		# Convert to GIF
		# https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html?highlight=duration#saving
		# duration := display duration of each frame in ms
		duration = int(1000 / framerate)
		frames_to_append = frames[1:] if len(frames) > 1 else []
		frames[0].save(dest_filename,
						format="GIF",
						append_images=frames_to_append,
						save_all=True,
						duration=duration,
						loop=0)

	def convert(self):
		'''Convert all images in the frames directory into a gif file.

		This function doesn't remove the frames after conversion.
		'''
		return self.convert_to_gif(self.filename, self.path_to_frames, self.framerate, num_files_to_convert=self.num_frames)

	def close(self):
		'''Close the animation, create the final gif file and (potentially) remove the individual frames.
		'''
		try:
			if not self.closed:
				self.convert()
				if self.cleanup:
					shutil.rmtree(self.path_to_frames, ignore_errors=True)
		finally:
			self.num_frames = 0
			self.closed = True

class FFMpegWriter(object):
	'''A writer of video files from Matplotlib figures.

	This class uses FFMpeg as the basis for writing frames to a video file.

	Parameters
	----------
	filename : string
		The filename of the generated video file.
	codec : string
		The codec for FFMpeg to use. If it is not given, a suitable codec
		will be guessed based on the file extension.
	framerate : integer
		The number of frames per second of the generated video file.
	quality : string
		The quality of the encoding for lossy codecs. Please refer to FFMpeg documentation.
	preset : string
		The preset for the quality of the encoding. Please refer to FFMpeg documentation.

	Raises
	------
	ValueError
		If the codec was not given and could not be guessed based on the file extension.
	RuntimeError
		If something went wrong during initialization of the call to FFMpeg. Most likely,
		FFMpeg is not installed and/or not available from the commandline.
	'''
	def __init__(self, filename, codec=None, framerate=24, quality=None, preset=None):
		if codec is None:
			extension = os.path.splitext(filename)[1]
			if extension == '.mp4' or extension == '.avi':
				codec = 'libx264'
			elif extension == '.webm':
				codec = 'libvpx-vp9'
			else:
				raise ValueError('No codec was given and it could not be guessed based on the file extension.')

		self.closed = True
		self.filename = filename
		self.codec = codec
		self.framerate = framerate

		if codec == 'libx264':
			if quality is None:
				quality = 10

			if preset is None:
				preset = 'veryslow'

			command = ['ffmpeg', '-y', '-nostats', '-v', 'quiet', '-f', 'image2pipe',
				'-vcodec','png', '-r', str(framerate), '-threads', '0', '-i', '-',
				'-vcodec', 'libx264', '-pix_fmt', 'yuv420p', '-preset', preset, '-r',
				str(framerate), '-crf', str(quality), filename]
		elif codec == 'mpeg4':
			if quality is None:
				quality = 4

			command = ['ffmpeg', '-y', '-nostats', '-v', 'quiet', '-f', 'image2pipe',
				'-vcodec','png', '-r', str(framerate), '-threads', '0', '-i', '-',
				'-vcodec', 'mpeg4', '-q:v', str(quality), '-r', str(framerate), filename]
		elif codec == 'libvpx-vp9':
			if quality is None:
				quality = 30

			command = ['ffmpeg', '-y', '-nostats', '-v', 'quiet', '-f', 'image2pipe',
				'-vcodec','png', '-r', str(framerate), '-threads', '0', '-i', '-']
			if quality < 0:
				command.extend(['-vcodec', 'libvpx-vp9', '-lossless', '1',
					'-r', str(framerate), filename])
			else:
				command.extend(['-vcodec', 'libvpx-vp9', '-crf', str(quality),
					'-b:v', '0', '-r', str(framerate), filename])
		else:
			raise ValueError('Codec unknown.')

		try:
			self.p = Popen(command, stdin=PIPE)
		except OSError:
			raise RuntimeError('Something went wrong when opening FFMpeg. Is FFMpeg installed and accessible from the command line?')
		self.closed = False

	def __del__(self):
		try:
			self.close()
		except Exception:
			import warnings
			warnings.warn('Something went wrong while closing FFMpeg...', RuntimeWarning)

	def add_frame(self, fig=None, data=None, cmap=None, dpi=None):
		'''Add a frame to the animation.

		Parameters
		----------
		fig : Matplotlib figure
			The Matplotlib figure acting as the animation frame.
		data : ndarray
			The image data array acting as the animation frame.
		cmap : Matplotlib colormap
			The optional colormap for the image data.
		dpi : integer or None
			The number of dots per inch with which to save the matplotlib figure.
			If it is not given, the default Matplotlib dpi will be used.

		Raises
		------
		RuntimeError
			If the function was called on a closed FFMpegWriter.
		'''
		if self.closed:
			raise RuntimeError('Attempted to add a frame to a closed FFMpegWriter.')

		if data is None:
			if fig is None:
				fig = matplotlib.pyplot.gcf()

			facecolor = list(fig.get_facecolor())
			facecolor[3] = 1

			fig.savefig(self.p.stdin, format='png', transparent=False, dpi=dpi, facecolor=facecolor)
		else:
			if cmap is not None:
				data = matplotlib.cm.get_cmap(cmap)(data, bytes=True)

			imageio.imwrite(self.p.stdin, data, format='png')

	def close(self):
		'''Close the animation writer and finish the video file.

		This closes the FFMpeg call.
		'''
		if not self.closed:
			self.p.stdin.close()
			self.p.wait()
			self.p = None
		self.closed = True

	def _repr_html_(self):
		'''Get an HTML representation of the generated video.

		Helper function for Jupyter notebooks. The video will be inline embedded in an
		HTML5 video tag using base64 encoding. This is not very efficient, so only use this
		for small video files.

		The FFMpegWriter must be closed for this function to work.

		Raises
		------
		RuntimeError
			If the call was made on an open FFMpegWriter.
		'''
		if not self.closed:
			raise RuntimeError('Attempted to show the generated movie on an opened FFMpegWriter.')

		video = open(self.filename, 'rb').read()
		video = base64.b64encode(video).decode('ascii').rstrip()

		if self.filename.endswith('.mp4'):
			mimetype = 'video/mp4'
		elif self.filename.endswith('.webm'):
			mimetype = 'video/webm'
		elif self.filename.endswith('.avi'):
			mimetype = 'video/avi'
		else:
			raise RuntimeError('Mimetype could not be guessed.')

		output = '''<video controls><source src="data:{0};base64,{1}" type="{0}">Your browser does not support the video tag.</video>'''
		output = output.format(mimetype, video)

		return output



import numpy as np
from .optical_element import OpticalElement, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward

class Apodizer(AgnosticOpticalElement):
	'''A thin apodizer.

	This apodizer can apodize both in phase and amplitude.

	Parameters
	----------
	apodization : Field or scalar or function of wavelength
		The apodization that we want to apply to any input wavefront.
	'''
	def __init__(self, apodization):
		self._apodization = apodization

		AgnosticOpticalElement.__init__(self, True, True)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		instance_data.apodization = self.evaluate_parameter(self.apodization, input_grid, output_grid, wavelength)

	@property
	def apodization(self):
		return self._apodization

	@apodization.setter
	def apodization(self, apodization):
		self._apodization = apodization

		self.clear_cache()

	def get_input_grid(self, output_grid, wavelength):
		return output_grid

	def get_output_grid(self, input_grid, wavelength):
		return input_grid

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		wf = wavefront.copy()
		wf.electric_field *= instance_data.apodization

		return wf

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		wf = wavefront.copy()
		wf.electric_field *= instance_data.apodization.conj()

		return wf

class PhaseApodizer(Apodizer):
	'''A phase-only thin apodizer.

	Parameters
	----------
	phase : Field or scalar or function
		The phase apodization.
	'''
	def __init__(self, phase):
		self._phase = phase

		Apodizer.__init__(self, self.apodization)

	@property
	def apodization(self):
		return self.construct_function(lambda p: np.exp(1j * p), self.phase)

	@property
	def phase(self):
		return self._phase

	@phase.setter
	def phase(self, phase):
		self._phase = phase

		self.clear_cache()

class SurfaceApodizer(Apodizer):
	'''A transmissive sagged surface optic.

	The surface is simulated as a thin plate. Propagation effects due to the
	thickness of the plate are not included. The supplied refractive index
	may change as function of wavelength.

	Parameters
	----------
	surface_sag : Field or scalar or function
		The sag in the surface.
	refractive_index : scalar or function
		The refractive index of the material of the plate.
	'''
	def __init__(self, surface_sag, refractive_index):
		self._surface_sag = surface_sag
		self._refractive_index = refractive_index

		Apodizer.__init__(self, self.apodization)

	@property
	def opd(self):
		return self.construct_function(lambda n, surf: (n - 1) * surf, self.refractive_index, self.surface_sag)

	optical_path_difference = opd

	@property
	def phase(self):
		return self.construct_function(lambda opd, wavelength: opd * 2 * np.pi / wavelength, self.opd)

	@property
	def apodization(self):
		return self.construct_function(lambda p: np.exp(1j * p), self.phase)

	@property
	def refractive_index(self):
		return self._refractive_index

	@refractive_index.setter
	def refractive_index(self, refractive_index):
		self._refractive_index = refractive_index

		self.clear_cache()

	@property
	def surface_sag(self):
		return self._surface_sag

	@surface_sag.setter
	def surface_sag(self, surface_sag):
		self._surface_sag = surface_sag

		self.clear_cache()

class ComplexSurfaceApodizer(OpticalElement):
	def __init__(self, amplitude, surface, refractive_index):
		self.amplitude = amplitude
		self.surface = surface
		self.refractive_index = refractive_index

	def phase_for(self, wavelength):
		'''Get the phase screen in radians at a certain wavelength.

		Parameters
		----------
		wavelength : scalar
			The wavelength at which to calculate the phase screen.
		'''
		wavenumber = 2 * np.pi / wavelength

		opd = (self.refractive_index - 1) * self.surface

		return opd * wavenumber

	def forward(self, wavefront):
		opd = (self.refractive_index(wavefront.wavelength) - 1) * self.surface

		wf = wavefront.copy()
		wf.electric_field *= self.amplitude * np.exp(1j * opd * wf.wavenumber)

		return wf

	def backward(self, wavefront):
		opd = (self.refractive_index(wavefront.wavelength) - 1) * self.surface

		wf = wavefront.copy()
		wf.electric_field *= self.amplitude * np.exp(-1j * opd * wf.wavenumber)

		return wf

class MultiplexedComplexSurfaceApodizer(OpticalElement):
	def __init__(self, amplitude, surface, refractive_index):
		self.amplitude = amplitude
		self.surface = surface
		self.refractive_index = refractive_index

	def forward(self, wavefront):
		apodizer_mask = 0
		for amplitude, surface in zip(self.amplitude, self.surface):
			opd = (self.refractive_index(wavefront.wavelength) - 1) * surface
			apodizer_mask += amplitude * np.exp(1j * opd * wavefront.wavenumber)

		wf = wavefront.copy()
		wf.electric_field *= apodizer_mask
		return wf

	def backward(self, wavefront):
		apodizer_mask = 0
		for amplitude, surface in zip(self.amplitude, self.surface):
			opd = (self.refractive_index(wavefront.wavelength) - 1) * surface
			apodizer_mask += amplitude * np.exp(1j * opd * wavefront.wavenumber)

		wf = wavefront.copy()
		wf.electric_field /= apodizer_mask
		return wf



from ..optics import Apodizer, OpticalElement, Wavefront
from ..field import Field
from ..fourier import FastFourierTransform, make_fft_grid

import numpy as np

try:
	import mkl_fft._numpy_fft as _fft_module
except ImportError:
	_fft_module = np.fft

class KnifeEdgeLyotCoronagraph(OpticalElement):
	'''A Lyot-style coronagraph with a centered, knife-edge focal-plane mask.

	As the knife-edge is invariant along one axis, the Fourier transforms can
	be sped up by only taking them along one axis. Also, the zero-padding only
	needs to be done along that axis, reducing memory usage while doing the
	transform.

	The knife-edge is set along the y-axis, so perpendicular to the x-axis.

	Parameters
	----------
	input_grid : Grid
		The grid of the wavefront that is to be propagated.
	q : scalar
		The amount of oversampling used for the knife edge.
	apodizer : OpticalElement or Field or None
		The pre-apodizer in the pupil before the focal-plane mask.
		If this is a Field, it will be converted into an apodizer.
		If this is None, no apodizer will be used.
	lyot_mask : OpticalElement or Field
		The Lyot stop in the pupil after the focal-plane mask. If
		this is a Field, it will be converted to an apodizer.
		If this is None, no Lyot stop will be used.
	'''
	def __init__(self, input_grid, q=8, apodizer=None, lyot_stop=None):
		fft = FastFourierTransform(input_grid, q, 1)
		self.cutout_input = (Ellipsis, fft.cutout_input[1])
		self.internal_shape = input_grid.shape[0], fft.internal_shape[1]

		# Create the knife-edge focal-plane mask along the x-axis.
		focal_mask_grid = make_fft_grid(input_grid, q).scaled(1.0 / (2 * np.pi))
		x = focal_mask_grid.separated_coords[0]
		self.focal_mask = (x > 0).astype('float')
		self.focal_mask[np.abs(x) < 1e-9] = 0.5

		# Pre-shift the mask to speed up propagations
		# FIXME: not necessary when 1D FFTs on 2D grids are implemented.
		self.focal_mask = np.fft.fftshift(self.focal_mask)

		if apodizer is not None and not hasattr(apodizer, 'input_grid'):
			apodizer = Apodizer(apodizer)
		self.apodizer = apodizer

		if lyot_stop is not None and not hasattr(lyot_stop, 'input_grid'):
			lyot_stop = Apodizer(lyot_stop)
		self.lyot_stop = lyot_stop

	def forward(self, wavefront):
		'''Propagate a wavefront forward through the knife-edge coronagraph.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		Wavefront
			The propagated wavefront.
		'''
		if self.apodizer is not None:
			wavefront = self.apodizer.forward(wavefront)

		ap = np.zeros(self.internal_shape, dtype='complex')
		ap[tuple(self.cutout_input)] = wavefront.electric_field.shaped

		# FIXME: use 1D FFTs on 2D grids implementation when available, instead of this.
		post_coro = _fft_module.ifft(_fft_module.fft(ap, axis=1) * self.focal_mask[np.newaxis,:], axis=1)
		post_coro = Field(post_coro[self.cutout_input].ravel(), wavefront.electric_field.grid)

		wavefront = Wavefront(post_coro, wavefront.wavelength)

		if self.lyot_stop is not None:
			wavefront = self.lyot_stop.forward(wavefront)

		return wavefront

	def backward(self, wavefront):
		'''Propagate a wavefront backward through the knife-edge coronagraph.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		Wavefront
			The propagated wavefront.
		'''
		if self.lyot_stop is not None:
			wavefront = self.lyot_stop.backward(wavefront)

		ap = np.zeros(self.internal_shape, dtype='complex')
		ap[tuple(self.cutout_input)] = wavefront.electric_field.shaped

		# FIXME: use 1D FFTs on 2D grids implementation when available, instead of this.
		post_coro = _fft_module.ifft(_fft_module.fft(ap, axis=1) * self.focal_mask[np.newaxis,:], axis=1)
		post_coro = Field(post_coro[self.cutout_input].ravel(), wavefront.electric_field.grid)

		wavefront = Wavefront(post_coro, wavefront.wavelength)

		if self.apodizer is not None:
			wavefront = self.apodizer.backward(wavefront)

		return wavefront



import numpy as np

def read_fits(filename):
	'''Read an array from a fits file.

	Parameters
	----------
	filename : string
		The filename of the file to read. This can include a path.

	Returns
	-------
	ndarray
		The ndarray read from the fits file.
	'''
	from astropy.io import fits
	return fits.getdata(filename).copy()

def write_fits(data, filename, shape=None, overwrite=True):
	'''Write the data to a fits-file.

	If the data is a Field with SeparatedCoords, the shaped field will be written to the
	fits-file, to allow for easier viewing with external tools.

	Parameters
	----------
	data : ndarray or Field
		The ndarray or Field to write to the fits file.
	filename : string
		The filename of the newly created file. This can include a path.
	shape : ndarray or None
		The shape to which to reshape the data. If this is given, it will override a potential
		shape from the grid accompaning the field.
	overwrite : boolean
		Whether to overwrite the fits-file if it already exists.
	'''
	from astropy.io import fits

	hdu = None

	if shape is not None:
		hdu = fits.PrimaryHDU(data.reshape(shape))
	elif hasattr(data, 'grid'):
		if data.grid.is_separated:
			hdu = fits.PrimaryHDU(data.shaped)

	if hdu is None:
		hdu = fits.PrimaryHDU(data)

	hdu.writeto(filename, overwrite=True)

def _guess_file_format(filename):
	'''Guess the file format from the filename.

	Parameters
	----------
	filename : string
		The filename of the file.

	Returns
	-------
	string
		The file type. This is one of ["asdf", "fits"].
	'''
	if filename.endswith('asdf'):
		return 'asdf'
	elif filename.endswith('fits') or filename.endswith('fits.gz'):
		return 'fits'
	elif filename.endswith('pkl') or filename.endswith('pickle'):
		return 'pickle'
	else:
		return None

def _make_metadata(file_type):
	'''Make a metadata section for file output.

	This contains the version of HCIPy used to write the file, the current time and
	date and the file type.

	Parameters
	----------
	file_type : string
		The identifier string for what is going to be stored in the file.

	Returns
	-------
	dictionary
		The metadata.
	'''
	from ..version import get_version
	import datetime

	tree = {
		'meta': {
			'author': 'HCIPy %s' % get_version(),
			'date_utc': datetime.datetime.utcnow().isoformat(),
			'file_type': file_type
		}
	}

	return tree

def read_grid(filename, fmt=None):
	'''Read a grid from a file.

	Parameters
	----------
	filename : string
		The path of the file you want to read the grid from.
	fmt : string
		The file format. If it is not given, the file format will be guessed from the file extension.

	Returns
	-------
	Grid
		The read grid.

	Raises
	------
	ValueError
		If the file format could not be guessed from the file extension.
	NotImplementedError
		If the file format was not yet implemented.
	'''
	from ..field import Grid

	if fmt is None:
		fmt = _guess_file_format(filename)

		if fmt is None:
			raise ValueError('Format not given and could not be guessed based on the file extension.')

	if fmt in ['asdf', 'fits']:
		import asdf

		f = asdf.open(filename)
		grid = Grid.from_dict(f.tree['grid'])
		f.close()

		return grid
	elif fmt == 'pickle':
		import pickle

		with open(filename, 'rb') as f:
			return pickle.load(f)
	else:
		raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def write_grid(grid, filename, fmt=None, overwrite=True):
	'''Write a grid to a file.

	Parameters
	----------
	grid : Grid
		The grid to write to the file.
	filename : string
		The path of the file you want to write to.
	fmt : string
		The file format. If it is not given, the file format will be guessed based on the file extension.
	overwrite : boolean
		Whether to overwrite the file, if it already exists.

	Raises
	------
	NotImplementedError
		If the file format was not yet implemented.
	'''
	if fmt is None:
		fmt = _guess_file_format(filename)

		if fmt is None:
			raise ValueError('Format not given and could not be guessed based on the file extension.')

	tree = _make_metadata('grid')
	tree['grid'] = grid.to_dict()

	if fmt == 'asdf':
		import asdf

		target = asdf.AsdfFile(tree)
		target.write_to(filename, all_array_compression='zlib')
	elif fmt == 'fits':
		from astropy.io import fits
		import asdf

		hdulist = fits.HDUList()

		ff = asdf.fits_embed.AsdfInFits(hdulist, tree)
		ff.write_to(filename, all_array_compression='zlib', overwrite=overwrite)
	elif fmt == 'pickle':
		import pickle

		with open(filename, 'wb') as f:
			pickle.dump(grid, f)
	else:
		raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def read_field(filename, fmt=None):
	'''Read a field from a file.

	Parameters
	----------
	filename : string
		The path of the file you want to read the field from.
	fmt : string
		The file format. If it is not given, the file format will be guessed from the file extension.

	Returns
	-------
	Field
		The read field.

	Raises
	------
	ValueError
		If the file format could not be guessed from the file extension.
	NotImplementedError
		If the file format was not yet implemented.
	'''
	from ..field import Field

	if fmt is None:
		fmt = _guess_file_format(filename)

		if fmt is None:
			raise ValueError('Format not given and could not be guessed based on the file extension.')

	if fmt == 'asdf':
		import asdf

		f = asdf.open(filename)
		field = Field.from_dict(f.tree['field'])
		f.close()

		return field
	elif fmt == 'fits':
		import asdf
		from ..field import Grid

		f = asdf.open(filename)
		tree = f.tree['field']

		if 'grid' in tree:
			grid = Grid.from_dict(tree['grid'])
			new_shape = np.concatenate((tree['values'].shape[:-grid.ndim], [grid.size])).astype('int')
			tree['values'] = tree['values'].reshape(new_shape)

		field = Field.from_dict(tree).reshape(new_shape)
		f.close()

		return field
	elif fmt == 'pickle':
		import pickle

		with open(filename, 'rb') as f:
			return pickle.load(f)
	else:
		raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def write_field(field, filename, fmt=None, overwrite=True):
	'''Write a field to a file.

	Parameters
	----------
	field : Field
		The field to write to the file.
	filename : string
		The path of the file you want to write to.
	fmt : string
		The file format. If it is not given, the file format will be guessed based on the file extension.
	overwrite : boolean
		Whether to overwrite the file, if it already exists.

	Raises
	------
	NotImplementedError
		If the file format was not yet implemented.
	'''
	if fmt is None:
		fmt = _guess_file_format(filename)

		if fmt is None:
			raise ValueError('Format not given and could not be guessed based on the file extension.')

	tree = _make_metadata('field')
	tree['field'] = field.to_dict()

	if fmt == 'asdf':
		import asdf

		target = asdf.AsdfFile(tree)
		target.write_to(filename, all_array_compression='zlib')
	elif fmt == 'fits':
		from astropy.io import fits
		import asdf

		hdulist = fits.HDUList()

		if field.grid.is_separated:
			hdulist.append(fits.ImageHDU(np.ascontiguousarray(field.shaped)))
			tree['field']['values'] = hdulist[0].data

		if field.grid.is_regular:
			from astropy import wcs

			w = wcs.WCS(naxis=field.grid.ndim)
			w.wcs.crpix = np.ones(field.grid.ndim)
			w.wcs.cdelt = field.grid.delta
			w.wcs.crval = field.grid.zero
			w.wcs.ctype = ['X', 'Y', 'Z', 'W'][:field.grid.ndim]

			hdulist[0].header.update(w.to_header())

		ff = asdf.fits_embed.AsdfInFits(hdulist, tree)
		ff.write_to(filename, all_array_compression='zlib', overwrite=overwrite)
	elif fmt == 'pickle':
		import pickle

		with open(filename, 'wb') as f:
			pickle.dump(field, f)
	else:
		raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def read_mode_basis(filename, fmt=None):
	'''Read a mode basis from a file.

	Parameters
	----------
	filename : string
		The path of the file you want to read the mode basis from.
	fmt : string
		The file format. If it is not given, the file format will be guessed from the file extension.

	Returns
	-------
	ModeBasis
		The read mode basis.

	Raises
	------
	ValueError
		If the file format could not be guessed from the file extension.
	NotImplementedError
		If the file format was not yet implemented.
	'''
	from ..mode_basis import ModeBasis

	if fmt is None:
		fmt = _guess_file_format(filename)

		if fmt is None:
			raise ValueError('Format not given and could not be guessed based on the file extension.')

	if fmt == 'asdf':
		import asdf

		f = asdf.open(filename)
		mode_basis = ModeBasis.from_dict(f.tree['mode_basis'])
		f.close()

		return mode_basis
	elif fmt == 'fits':
		import asdf
		from ..field import Grid

		f = asdf.open(filename)
		tree = f.tree['mode_basis']

		if 'modes' in tree:
			grid = Grid.from_dict(tree['grid'])
			modes = tree['modes']

			old_shape = np.concatenate((modes.shape[:-grid.ndim], [grid.size]))
			tree['transformation_matrix'] = modes.reshape(old_shape).T
			del tree['modes']

		mode_basis = ModeBasis.from_dict(tree)
		f.close()

		return mode_basis
	elif fmt == 'pickle':
		import pickle

		with open(filename, 'rb') as f:
			return pickle.load(f)
	else:
		raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)

def write_mode_basis(mode_basis, filename, fmt=None, overwrite=True):
	'''Write a mode basis to a file.

	Parameters
	----------
	mode_basis : ModeBasis
		The mode basis to write to the file.
	filename : string
		The path of the file you want to write to.
	fmt : string
		The file format. If it is not given, the file format will be guessed based on the file extension.
	overwrite : boolean
		Whether to overwrite the file, if it already exists.

	Raises
	------
	NotImplementedError
		If the file format was not yet implemented.
	'''
	if fmt is None:
		fmt = _guess_file_format(filename)

		if fmt is None:
			raise ValueError('Format not given and could not be guessed based on the file extension.')

	tree = _make_metadata('mode_basis')
	tree['mode_basis'] = mode_basis.to_dict()

	if fmt == 'asdf':
		import asdf

		target = asdf.AsdfFile(tree)
		target.write_to(filename, all_array_compression='zlib')
	elif fmt == 'fits':
		from astropy.io import fits
		import asdf
		from astropy import wcs

		hdulist = fits.HDUList()

		if mode_basis.grid and mode_basis.grid.is_separated:
			new_shape = np.concatenate(([-1], mode_basis.grid.shape))
			modes = np.ascontiguousarray(mode_basis.to_dense().transformation_matrix.T.reshape(new_shape))

			hdulist.append(fits.ImageHDU(np.asarray(modes)))

			w = wcs.WCS(naxis=mode_basis.grid.ndim)
			w.wcs.crpix = np.ones(mode_basis.grid.ndim)
			w.wcs.cdelt = mode_basis.grid.delta
			w.wcs.crval = mode_basis.grid.zero
			w.wcs.ctype = ['X', 'Y', 'Z', 'W'][:mode_basis.grid.ndim]

			hdulist[0].header.update(w.to_header())

			del tree['mode_basis']['transformation_matrix']
			tree['mode_basis']['modes'] = hdulist[0].data

		ff = asdf.fits_embed.AsdfInFits(hdulist, tree)
		ff.write_to(filename, all_array_compression='zlib', overwrite=overwrite)
	elif fmt == 'pickle':
		import pickle

		with open(filename, 'wb') as f:
			pickle.dump(mode_basis, f)
	else:
		raise NotImplementedError('The "%s" file format has not been implemented.' % fmt)



import numpy as np
import scipy.sparse
import scipy.sparse.linalg

class SVD(object):
	'''The Singular Value Decomposition for the provided matrix.

	This class wraps two versions of the SVD in numpy and scipy, and provides
	easy access to singular modes (as mode bases) and allows for calculation
	of the SVD for a limited number of modes.

	When a sparse matrix is passed, and no number of modes is given, all but one
	mode will be calculated. The reason is that the sparse SVD implementation in Scipy
	doesn't allow calculation of all modes. If all modes are required, the user
	must pass a densified version of the matrix (ie. `M.toarray()`).

	Parameters
	----------
	M : ndarray or any sparse matrix
		The matrix on which to perform the SVD.
	num_modes : int or None
		The number of singular values and modes to calculate. If this is None,
		and `M` is not sparse, all modes will be computed. If this is None and
		`M` is sparse, all but one mode will be computed.
	'''
	def __init__(self, M, num_modes=None):
		self._M = M
		self._num_modes = num_modes

		is_sparse = scipy.sparse.issparse(M)

		if is_sparse and self.num_modes is None:
			self._num_modes = min(M.shape) - 1

		if self.num_modes is None:
			self._svd = np.linalg.svd(M, full_matrices=False)
		else:
			self._svd = scipy.sparse.linalg.svds(M, int(self.num_modes))

	@property
	def left_singular_modes(self):
		'''The left singular modes of the matrix, as a ModeBasis.
		'''
		from ..mode_basis import ModeBasis

		return ModeBasis([m for m in self.U.conj().T])

	@property
	def right_singular_modes(self):
		'''The right singular modes of the matrix, as a ModeBasis.
		'''
		from ..mode_basis import ModeBasis

		return ModeBasis([m for m in self.Vt.conj()])

	@property
	def singular_values(self):
		'''The singular values of the matrix.
		'''
		return self.S

	@property
	def U(self):
		'''The U matrix of the SVD.
		'''
		return self.svd[0]

	@property
	def S(self):
		'''The singular values of the matrix.
		'''
		return self.svd[1]

	@property
	def Vt(self):
		'''The V^T matrix of the SVD.
		'''
		return self.svd[2]

	def __getitem__(self, i):
		'''The raw U, S, and V^T matrices of the SVD by index.
		'''
		return self.svd[i]

	@property
	def svd(self):
		'''The raw U, S, and V^T matrices of the SVD as a tuple.
		'''
		return self._svd

	@property
	def num_modes(self):
		'''The number of singular modes that were calculated in this SVD.
		'''
		return self._num_modes

	@property
	def M(self):
		'''The matrix for which the SVD was calculated.
		'''
		return self._M
